# Licensed under a 3-clause BSD style license - see LICENSE.rst

from copy import deepcopy

import numpy as np

from astropy import units as u
from astropy.table import QTable, Table, groups
from astropy.time import Time, TimeDelta
from astropy.timeseries.core import BaseTimeSeries, autocheck_required_columns
from astropy.units import Quantity

__all__ = ["BinnedTimeSeries"]


@autocheck_required_columns
class BinnedTimeSeries(BaseTimeSeries):
    """
    A class to represent binned time series data in tabular form.

    `~astropy.timeseries.BinnedTimeSeries` provides a class for
    representing time series as a collection of values of different
    quantities measured in time bins (for time series with values
    sampled at specific times, see the `~astropy.timeseries.TimeSeries`
    class). `~astropy.timeseries.BinnedTimeSeries` is a sub-class of
    `~astropy.table.QTable` and thus provides all the standard table
    maniplation methods available to tables, but it also provides
    additional conveniences for dealing with time series, such as a
    flexible initializer for setting up the times, and attributes to
    access the start/center/end time of bins.

    See also: https://docs.astropy.org/en/stable/timeseries/

    Parameters
    ----------
    data : numpy ndarray, dict, list, table-like object, optional
        Data to initialize time series. This does not need to contain the
        times, which can be provided separately, but if it does contain the
        times they should be in columns called ``'time_bin_start'`` and
        ``'time_bin_size'`` to be automatically recognized.
    time_bin_start : `~astropy.time.Time` or iterable
        The times of the start of each bin - this can be either given
        directly as a `~astropy.time.Time` array or as any iterable that
        initializes the `~astropy.time.Time` class. If this is given, then
        the remaining time-related arguments should not be used. This can also
        be a scalar value if ``time_bin_size`` is provided.
    time_bin_end : `~astropy.time.Time` or iterable
        The times of the end of each bin - this can be either given directly
        as a `~astropy.time.Time` array or as any value or iterable that
        initializes the `~astropy.time.Time` class. If this is given, then the
        remaining time-related arguments should not be used. This can only be
        given if ``time_bin_start`` is an array of values. If ``time_bin_end``
        is a scalar, time bins are assumed to be contiguous, such that the end
        of each bin is the start of the next one, and ``time_bin_end`` gives
        the end time for the last bin. If ``time_bin_end`` is an array, the
        time bins do not need to be contiguous. If this argument is provided,
        ``time_bin_size`` should not be provided.
    time_bin_size : `~astropy.time.TimeDelta` or `~astropy.units.Quantity`
        The size of the time bins, either as a scalar value (in which case all
        time bins will be assumed to have the same duration) or as an array of
        values (in which case each time bin can have a different duration).
        If this argument is provided, ``time_bin_end`` should not be provided.
    n_bins : int
        The number of time bins for the series. This is only used if both
        ``time_bin_start`` and ``time_bin_size`` are provided and are scalar
        values.
    **kwargs : dict, optional
        Additional keyword arguments are passed to `~astropy.table.QTable`.
    """

    _required_columns = ["time_bin_start", "time_bin_size"]

    def __init__(
        self,
        data=None,
        *,
        time_bin_start=None,
        time_bin_end=None,
        time_bin_size=None,
        n_bins=None,
        **kwargs,
    ):
        super().__init__(data=data, **kwargs)

        # For some operations, an empty time series needs to be created, then
        # columns added one by one. We should check that when columns are added
        # manually, time is added first and is of the right type.
        if (
            data is None
            and time_bin_start is None
            and time_bin_end is None
            and time_bin_size is None
            and n_bins is None
        ):
            self._required_columns_relax = True
            return

        # First if time_bin_start and time_bin_end have been given in the table data, we
        # should extract them and treat them as if they had been passed as
        # keyword arguments.

        if "time_bin_start" in self.colnames:
            if time_bin_start is None:
                time_bin_start = self.columns["time_bin_start"]
            else:
                raise TypeError(
                    "'time_bin_start' has been given both in the table "
                    "and as a keyword argument"
                )

        if "time_bin_size" in self.colnames:
            if time_bin_size is None:
                time_bin_size = self.columns["time_bin_size"]
            else:
                raise TypeError(
                    "'time_bin_size' has been given both in the table "
                    "and as a keyword argument"
                )

        if time_bin_start is None:
            raise TypeError("'time_bin_start' has not been specified")

        if time_bin_end is None and time_bin_size is None:
            raise TypeError(
                "Either 'time_bin_size' or 'time_bin_end' should be specified"
            )

        if not isinstance(time_bin_start, (Time, TimeDelta)):
            time_bin_start = Time(time_bin_start)

        if time_bin_end is not None and not isinstance(time_bin_end, (Time, TimeDelta)):
            time_bin_end = Time(time_bin_end)

        if time_bin_size is not None and not isinstance(
            time_bin_size, (Quantity, TimeDelta)
        ):
            raise TypeError("'time_bin_size' should be a Quantity or a TimeDelta")

        if isinstance(time_bin_size, TimeDelta):
            time_bin_size = time_bin_size.sec * u.s

        if n_bins is not None and time_bin_size is not None:
            if not (time_bin_start.isscalar and time_bin_size.isscalar):
                raise TypeError(
                    "'n_bins' cannot be specified if 'time_bin_start' or "
                    "'time_bin_size' are not scalar'"
                )

        if time_bin_start.isscalar:
            # We interpret this as meaning that this is the start of the
            # first bin and that the bins are contiguous. In this case,
            # we require time_bin_size to be specified.

            if time_bin_size is None:
                raise TypeError(
                    "'time_bin_start' is scalar, so 'time_bin_size' is required"
                )

            if time_bin_size.isscalar:
                if data is not None:
                    if n_bins is not None:
                        if n_bins != len(self):
                            raise TypeError(
                                "'n_bins' has been given and it is not the "
                                "same length as the input data."
                            )
                    else:
                        n_bins = len(self)

                time_bin_size = np.repeat(time_bin_size, n_bins)

            time_delta = np.cumsum(time_bin_size)
            time_bin_end = time_bin_start + time_delta

            # Now shift the array so that the first entry is 0
            time_delta = np.roll(time_delta, 1)
            time_delta[0] = 0.0 * u.s

            # Make time_bin_start into an array
            time_bin_start = time_bin_start + time_delta

        else:
            if len(self.colnames) > 0 and len(time_bin_start) != len(self):
                raise ValueError(
                    f"Length of 'time_bin_start' ({len(time_bin_start)}) should match "
                    f"table length ({len(self)})"
                )

            if time_bin_end is not None:
                if time_bin_end.isscalar:
                    times = time_bin_start.copy()
                    times[:-1] = times[1:]
                    times[-1] = time_bin_end
                    time_bin_end = times
                time_bin_size = (time_bin_end - time_bin_start).sec * u.s

        if time_bin_size.isscalar:
            time_bin_size = np.repeat(time_bin_size, len(self))

        with self._delay_required_column_checks():
            if "time_bin_start" in self.colnames:
                self.remove_column("time_bin_start")

            if "time_bin_size" in self.colnames:
                self.remove_column("time_bin_size")

            self.add_column(time_bin_start, index=0, name="time_bin_start")
            self.add_index("time_bin_start")
            self.add_column(time_bin_size, index=1, name="time_bin_size")

    @property
    def time_bin_start(self):
        """
        The start times of all the time bins.
        """
        return self["time_bin_start"]

    @property
    def time_bin_center(self):
        """
        The center times of all the time bins.
        """
        return self["time_bin_start"] + self["time_bin_size"] * 0.5

    @property
    def time_bin_end(self):
        """
        The end times of all the time bins.
        """
        return self["time_bin_start"] + self["time_bin_size"]

    @property
    def time_bin_size(self):
        """
        The sizes of all the time bins.
        """
        return self["time_bin_size"]

    def __getitem__(self, item):
        if self._is_list_or_tuple_of_str(item):
            if "time_bin_start" not in item or "time_bin_size" not in item:
                out = QTable(
                    [self[x] for x in item],
                    meta=deepcopy(self.meta),
                    copy_indices=self._copy_indices,
                )
                out._groups = groups.TableGroups(
                    out, indices=self.groups._indices, keys=self.groups._keys
                )
                return out
        return super().__getitem__(item)

    @classmethod
    def read(
        cls,
        filename,
        time_bin_start_column=None,
        time_bin_end_column=None,
        time_bin_size_column=None,
        time_bin_size_unit=None,
        time_format=None,
        time_scale=None,
        format=None,
        *args,
        **kwargs,
    ):
        """
        Read and parse a file and returns a `astropy.timeseries.BinnedTimeSeries`.

        This method uses the unified I/O infrastructure in Astropy which makes
        it easy to define readers/writers for various classes
        (https://docs.astropy.org/en/stable/io/unified.html). By default, this
        method will try and use readers defined specifically for the
        `astropy.timeseries.BinnedTimeSeries` class - however, it is also
        possible to use the ``format`` keyword to specify formats defined for
        the `astropy.table.Table` class - in this case, you will need to also
        provide the column names for column containing the start times for the
        bins, as well as other column names (see the Parameters section below
        for details)::

            >>> from astropy.timeseries.binned import BinnedTimeSeries
            >>> ts = BinnedTimeSeries.read('binned.dat', format='ascii.ecsv',
            ...                            time_bin_start_column='date_start',
            ...                            time_bin_end_column='date_end')  # doctest: +SKIP

        Parameters
        ----------
        filename : str
            File to parse.
        format : str
            File format specifier.
        time_bin_start_column : str
            The name of the column with the start time for each bin.
        time_bin_end_column : str, optional
            The name of the column with the end time for each bin. Either this
            option or ``time_bin_size_column`` should be specified.
        time_bin_size_column : str, optional
            The name of the column with the size for each bin. Either this
            option or ``time_bin_end_column`` should be specified.
        time_bin_size_unit : `astropy.units.Unit`, optional
            If ``time_bin_size_column`` is specified but does not have a unit
            set in the table, you can specify the unit manually.
        time_format : str, optional
            The time format for the start and end columns.
        time_scale : str, optional
            The time scale for the start and end columns.
        *args : tuple, optional
            Positional arguments passed through to the data reader.
        **kwargs : dict, optional
            Keyword arguments passed through to the data reader.

        Returns
        -------
        out : `astropy.timeseries.binned.BinnedTimeSeries`
            BinnedTimeSeries corresponding to the file.

        """
        try:
            # First we try the readers defined for the BinnedTimeSeries class
            return super().read(filename, *args, format=format, **kwargs)

        except TypeError:
            # Otherwise we fall back to the default Table readers

            if time_bin_start_column is None:
                raise ValueError(
                    "``time_bin_start_column`` should be provided since the default"
                    " Table readers are being used."
                )
            if time_bin_end_column is None and time_bin_size_column is None:
                raise ValueError(
                    "Either `time_bin_end_column` or `time_bin_size_column` should be"
                    " provided."
                )
            elif time_bin_end_column is not None and time_bin_size_column is not None:
                raise ValueError(
                    "Cannot specify both `time_bin_end_column` and"
                    " `time_bin_size_column`."
                )

            table = Table.read(filename, *args, format=format, **kwargs)

            if time_bin_start_column in table.colnames:
                time_bin_start = Time(
                    table.columns[time_bin_start_column],
                    scale=time_scale,
                    format=time_format,
                )
                table.remove_column(time_bin_start_column)
            else:
                raise ValueError(
                    f"Bin start time column '{time_bin_start_column}' not found in the"
                    " input data."
                )

            if time_bin_end_column is not None:
                if time_bin_end_column in table.colnames:
                    time_bin_end = Time(
                        table.columns[time_bin_end_column],
                        scale=time_scale,
                        format=time_format,
                    )
                    table.remove_column(time_bin_end_column)
                else:
                    raise ValueError(
                        f"Bin end time column '{time_bin_end_column}' not found in the"
                        " input data."
                    )

                time_bin_size = None

            elif time_bin_size_column is not None:
                if time_bin_size_column in table.colnames:
                    time_bin_size = table.columns[time_bin_size_column]
                    table.remove_column(time_bin_size_column)
                else:
                    raise ValueError(
                        f"Bin size column '{time_bin_size_column}' not found in the"
                        " input data."
                    )

                if time_bin_size.unit is None:
                    if time_bin_size_unit is None or not isinstance(
                        time_bin_size_unit, u.UnitBase
                    ):
                        raise ValueError(
                            "The bin size unit should be specified as an astropy Unit"
                            " using ``time_bin_size_unit``."
                        )
                    time_bin_size = time_bin_size * time_bin_size_unit
                else:
                    time_bin_size = u.Quantity(time_bin_size)

                time_bin_end = None

            if time_bin_start.isscalar and time_bin_size.isscalar:
                return cls(
                    data=table,
                    time_bin_start=time_bin_start,
                    time_bin_end=time_bin_end,
                    time_bin_size=time_bin_size,
                    n_bins=len(table),
                )
            else:
                return cls(
                    data=table,
                    time_bin_start=time_bin_start,
                    time_bin_end=time_bin_end,
                    time_bin_size=time_bin_size,
                )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Colloquially used Imperial units.

These units are available in the `astropy.units.imperial` namespace, but not in the
top-level `astropy.units` namespace, e.g.::

    >>> import astropy.units as u
    >>> mph = u.imperial.mile / u.hour
    >>> mph
    Unit("mi / h")

To include them in `~astropy.units.UnitBase.compose` and the results of
`~astropy.units.UnitBase.find_equivalent_units`, do::

    >>> import astropy.units as u
    >>> u.imperial.enable()  # doctest: +SKIP
"""
# avoid ruff complaints about undefined names defined by def_unit
# ruff: noqa: F821

__all__: list[str] = ["enable"]  #  Units are added at the end

from . import si
from .core import _UnitContext, add_enabled_units, def_unit
from .docgen import generate_dunder_all, generate_unit_summary

_ns = globals()

###########################################################################
# LENGTH

def_unit(["inch"], 2.54 * si.cm, namespace=_ns, doc="International inch")
def_unit(["ft", "foot"], 12 * inch, namespace=_ns, doc="International foot")
def_unit(["yd", "yard"], 3 * ft, namespace=_ns, doc="International yard")
def_unit(["mi", "mile"], 5280 * ft, namespace=_ns, doc="International mile")
def_unit(["mil", "thou"], 0.001 * inch, namespace=_ns, doc="Thousandth of an inch")
def_unit(["nmi", "nauticalmile", "NM"], 1852 * si.m, namespace=_ns, doc="Nautical mile")
def_unit(["fur", "furlong"], 660 * ft, namespace=_ns, doc="Furlong")


###########################################################################
# AREAS

def_unit(["ac", "acre"], 43560 * ft**2, namespace=_ns, doc="International acre")


###########################################################################
# VOLUMES

def_unit(["gallon"], si.liter / 0.264172052, namespace=_ns, doc="U.S. liquid gallon")
def_unit(["quart"], gallon / 4, namespace=_ns, doc="U.S. liquid quart")
def_unit(["pint"], quart / 2, namespace=_ns, doc="U.S. liquid pint")
def_unit(["cup"], pint / 2, namespace=_ns, doc="U.S. customary cup")
def_unit(
    ["foz", "fluid_oz", "fluid_ounce"], cup / 8, namespace=_ns, doc="U.S. fluid ounce"
)
def_unit(
    ["tbsp", "tablespoon"], foz / 2, namespace=_ns, doc="U.S. customary tablespoon"
)
def_unit(["tsp", "teaspoon"], tbsp / 3, namespace=_ns, doc="U.S. customary teaspoon")


###########################################################################
# MASS

def_unit(
    ["oz", "ounce"],
    28.349523125 * si.g,
    namespace=_ns,
    doc="International avoirdupois ounce: mass",
)
def_unit(
    ["lb", "lbm", "pound"],
    16 * oz,
    namespace=_ns,
    doc="International avoirdupois pound: mass",
)
def_unit(
    ["st", "stone"], 14 * lb, namespace=_ns, doc="International avoirdupois stone: mass"
)
def_unit(["ton"], 2000 * lb, namespace=_ns, doc="International avoirdupois ton: mass")
def_unit(["slug"], 32.174049 * lb, namespace=_ns, doc="slug: mass")


###########################################################################
# SPEED

def_unit(
    ["kn", "kt", "knot", "NMPH"],
    nmi / si.h,
    namespace=_ns,
    doc="nautical unit of speed: 1 nmi per hour",
)


###########################################################################
# FORCE

def_unit("lbf", slug * ft * si.s**-2, namespace=_ns, doc="Pound: force")
def_unit(["kip", "kilopound"], 1000 * lbf, namespace=_ns, doc="Kilopound: force")


##########################################################################
# ENERGY

def_unit(["BTU", "btu"], 1.05505585 * si.kJ, namespace=_ns, doc="British thermal unit")
def_unit(
    ["cal", "calorie"],
    4.184 * si.J,
    namespace=_ns,
    doc="Thermochemical calorie: pre-SI metric unit of energy",
)
def_unit(
    ["kcal", "Cal", "Calorie", "kilocal", "kilocalorie"],
    1000 * cal,
    namespace=_ns,
    doc="Calorie: colloquial definition of Calorie",
)


##########################################################################
# PRESSURE

def_unit("psi", lbf * inch**-2, namespace=_ns, doc="Pound per square inch: pressure")


###########################################################################
# POWER

# Imperial units
def_unit(
    ["hp", "horsepower"],
    si.W / 0.00134102209,
    namespace=_ns,
    doc="Electrical horsepower",
)


###########################################################################
# TEMPERATURE

def_unit(
    ["deg_F", "Fahrenheit"],
    namespace=_ns,
    doc="Degrees Fahrenheit",
    format={"latex": r"{}^{\circ}F", "unicode": "°F"},
)
def_unit(
    ["deg_R", "Rankine"],
    5 / 9 * si.K,
    namespace=_ns,
    doc="Rankine scale: absolute scale of thermodynamic temperature",
    format={"latex": r"{}^{\circ}R", "unicode": "°R"},
)

###########################################################################
# ALL & DOCSTRING

__all__ += generate_dunder_all(globals())  # noqa: PLE0605

if __doc__ is not None:
    # This generates a docstring for this module that describes all of the
    # standard units defined here.
    __doc__ += generate_unit_summary(globals())


def enable() -> _UnitContext:
    """
    Enable Imperial units so they appear in results of
    `~astropy.units.UnitBase.find_equivalent_units` and
    `~astropy.units.UnitBase.compose`.

    This may be used with the ``with`` statement to enable Imperial
    units only temporarily.
    """
    return add_enabled_units(globals())



# Licensed under a 3-clause BSD style license - see PYFITS.rst

import os
import pathlib
import shutil
import stat
import tempfile
import time

import pytest

from astropy.io import fits


@pytest.fixture(
    params=[False, "str", "pathlib"], ids=["", "home_is_data", "home_is_data, pathlib"]
)
def home_is_data(request, monkeypatch):
    """
    Pytest fixture to run a test case both with and without tilde paths.

    In the tilde-path case, calls like self.data('filename.fits') will
    produce '~/filename.fits', and environment variables will be temporarily
    modified so that '~' resolves to the data directory.
    """
    # This checks the value specified in the fixture annotation
    if request.param:
        # `request.instance` refers to the test case that's using this fixture.
        request.instance.monkeypatch = monkeypatch
        request.instance.set_home_as_data()

        request.instance.set_paths_via_pathlib(request.param == "pathlib")


@pytest.fixture(
    params=[False, "str", "pathlib"], ids=["", "home_is_data", "home_is_data, pathlib"]
)
def home_is_temp(request, monkeypatch):
    """
    Pytest fixture to run a test case both with and without tilde paths.

    In the tilde-path case, calls like self.temp('filename.fits') will
    produce '~/filename.fits', and environment variables will be temporarily
    modified so that '~' resolves to the temp directory. These files will also
    be tracked so that, after the test case, we can verify no files were written
    to a literal tilde path.
    """
    # This checks the value specified in the fixture annotation
    if request.param:
        # `request.instance` refers to the test case that's using this fixture.
        request.instance.monkeypatch = monkeypatch
        request.instance.set_home_as_temp()

        request.instance.set_paths_via_pathlib(request.param == "pathlib")


class FitsTestCase:
    def setup_method(self):
        self.data_dir = os.path.join(os.path.dirname(__file__), "data")
        self.temp_dir = tempfile.mkdtemp(prefix="fits-test-")

        self.home_is_data = False
        self.home_is_temp = False
        self.temp_files_used = set()
        self.use_pathlib = False

        # Restore global settings to defaults
        # TODO: Replace this when there's a better way to in the config API to
        # force config values to their defaults
        fits.conf.enable_record_valued_keyword_cards = True
        fits.conf.extension_name_case_sensitive = False
        fits.conf.strip_header_whitespace = True
        fits.conf.use_memmap = True

    def teardown_method(self):
        if self.home_is_temp:
            # Verify that no files were written to a literal tilde path
            for temp_file, temp_file_no_tilde in self.temp_files_used:
                assert not os.path.exists(temp_file)
                assert os.path.exists(temp_file_no_tilde)

        if hasattr(self, "temp_dir") and os.path.exists(self.temp_dir):
            tries = 3
            while tries:
                try:
                    shutil.rmtree(self.temp_dir)
                    break
                except OSError:
                    # Probably couldn't delete the file because for whatever
                    # reason a handle to it is still open/hasn't been
                    # garbage-collected
                    time.sleep(0.5)
                    tries -= 1

        fits.conf.reset("enable_record_valued_keyword_cards")
        fits.conf.reset("extension_name_case_sensitive")
        fits.conf.reset("strip_header_whitespace")
        fits.conf.reset("use_memmap")

    def copy_file(self, filename, new_filename=None):
        """Copies a backup of a test data file to the temp dir and sets its
        mode to writeable. Can optionally change the file's name.
        """
        new_filename = filename if new_filename is None else new_filename
        p = self.temp(new_filename)
        shutil.copy(
            os.path.expanduser(self.data(filename)),
            os.path.expanduser(p),
        )
        os.chmod(os.path.expanduser(p), stat.S_IREAD | stat.S_IWRITE)
        return p

    def data(self, filename):
        """Returns the path to a test data file."""
        if self.home_is_data:
            prefix = "~"
        else:
            prefix = self.data_dir

        if self.use_pathlib:
            return pathlib.Path(prefix, filename)
        return os.path.join(prefix, filename)

    def temp(self, filename):
        """Returns the full path to a file in the test temp dir."""
        real_target = os.path.join(self.temp_dir, filename)
        if self.home_is_temp:
            prefix = "~"
            # Record the '~' path and the intended path, for use
            # in `home_is_temp`
            self.temp_files_used.add((os.path.join(prefix, filename), real_target))
        else:
            prefix = self.temp_dir

        if self.use_pathlib:
            return pathlib.Path(prefix, filename)
        return os.path.join(prefix, filename)

    def set_home_as_data(self):
        """
        This overrides the HOME environment variable, so that paths beginning
        with '~/' expand to the data directory. Used by the `home_is_data`
        fixture.
        """
        self.home_is_data = True
        # For Unix
        self.monkeypatch.setenv("HOME", self.data_dir)
        # For Windows
        self.monkeypatch.setenv("USERPROFILE", self.data_dir)

    def set_home_as_temp(self):
        """
        This overrides the HOME environment variable, so that paths beginning
        with '~/' expand to the temp directory. In conjunction with
        self.temp(), temporary files are tracked as they are created, so we can
        verify they end up in the temporary directory and not unexpected places
        in the filesystem. Used by the `home_is_temp` fixture.
        """
        self.home_is_temp = True
        # For Unix
        self.monkeypatch.setenv("HOME", self.temp_dir)
        # For Windows
        self.monkeypatch.setenv("USERPROFILE", self.temp_dir)

    def set_paths_via_pathlib(self, use_pathlib):
        self.use_pathlib = use_pathlib



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Tests for custom error and warning messages in `astropy.coordinates`."""

from typing import NamedTuple

import pytest

from astropy import units as u
from astropy.coordinates import (
    GCRS,
    ICRS,
    BaseCoordinateFrame,
    Galactic,
    NonRotationTransformationError,
    NonRotationTransformationWarning,
    UnknownSiteException,
)


class FrameDescription(NamedTuple):
    frame: BaseCoordinateFrame
    description: str
    pytest_id: str


galactic = FrameDescription(
    Galactic(0 * u.deg, 0 * u.deg), "Galactic Frame", "Galactic"
)
gcrs_custom = FrameDescription(
    GCRS(
        0 * u.deg,
        0 * u.deg,
        obstime="J1950",
        obsgeovel=[30, -7, 11] * u.km / u.s,
    ),
    (
        "GCRS Frame (obstime=J1950.000, obsgeoloc=(0., 0., 0.) m, "
        "obsgeovel=(30000., -7000., 11000.) m / s)"
    ),
    "custom_GCRS",
)
gcrs_default = FrameDescription(
    GCRS(0 * u.deg, 0 * u.deg),
    (
        "GCRS Frame (obstime=J2000.000, obsgeoloc=(0., 0., 0.) m, "
        "obsgeovel=(0., 0., 0.) m / s)"
    ),
    "default_GCRS",
)
icrs = FrameDescription(ICRS(0 * u.deg, 0 * u.deg), "ICRS Frame", "ICRS")


@pytest.mark.parametrize(
    "coord_from,coord_to",
    [pytest.param(icrs, gcrs_custom), pytest.param(gcrs_default, galactic)],
    ids=lambda x: x.pytest_id,
)
def test_NonRotationTransformationError_message(coord_from, coord_to):
    assert str(NonRotationTransformationError(coord_to.frame, coord_from.frame)) == (
        f"refusing to transform other coordinates from <{coord_from.description}> to "
        f"<{coord_to.description}> because angular separation can depend on the "
        "direction of the transformation"
    )


@pytest.mark.parametrize(
    "coord_from,coord_to",
    [pytest.param(icrs, gcrs_default), pytest.param(gcrs_custom, galactic)],
    ids=lambda x: x.pytest_id,
)
def test_NonRotationTransformationWarning_message(coord_from, coord_to):
    assert str(NonRotationTransformationWarning(coord_to.frame, coord_from.frame)) == (
        f"transforming other coordinates from <{coord_from.description}> to "
        f"<{coord_to.description}>. Angular separation can depend on the direction of "
        "the transformation."
    )


@pytest.mark.parametrize(
    "site,attribute,message",
    [
        (
            "CERN",
            "EarthLocation.get_site_names",
            (
                "Site 'CERN' not in database. Use EarthLocation.get_site_names to "
                "see available sites. If 'CERN' exists in the online astropy-data "
                "repository, use the 'refresh_cache=True' option to download the "
                "latest version."
            ),
        ),
        (
            "Fermilab",
            "the 'names' attribute",
            (
                "Site 'Fermilab' not in database. Use the 'names' attribute to "
                "see available sites. If 'Fermilab' exists in the online astropy-data "
                "repository, use the 'refresh_cache=True' option to download the "
                "latest version."
            ),
        ),
    ],
)
def test_UnknownSiteException(site, attribute, message):
    assert str(UnknownSiteException(site, attribute)) == message


@pytest.mark.parametrize(
    "close_names,suggestion_str",
    [
        pytest.param(["greenwich"], "'greenwich'", id="one_suggestion"),
        pytest.param(
            ["greenwich", "Greenwich"], "'greenwich', 'Greenwich'", id="two_suggestions"
        ),
    ],
)
def test_UnknownSiteException_with_suggestions(close_names, suggestion_str):
    assert str(
        UnknownSiteException("grenwich", "EarthLocation.get_site_names", close_names)
    ) == (
        "Site 'grenwich' not in database. Use EarthLocation.get_site_names to see "
        "available sites. If 'grenwich' exists in the online astropy-data repository, "
        "use the 'refresh_cache=True' option to download the latest version. Did you "
        f"mean one of: {suggestion_str}?"
    )



"""Matter component."""

from numpy.typing import ArrayLike

from astropy.cosmology._src.typing import FArray
from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity

__all__ = ["MatterComponent"]


class MatterComponent:
    Om0: Quantity
    """Omega matter; matter density/critical density at z=0."""

    @deprecated_keywords("z", since="7.0")
    def Om(self, z: Quantity | ArrayLike) -> FArray:
        """Return the density parameter for non-relativistic matter at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Om : ndarray
            The density of non-relativistic matter relative to the critical
            density at each redshift.

        Notes
        -----
        This does not include neutrinos, even if non-relativistic at the
        redshift of interest.
        """
        z = aszarr(z)
        return self.Om0 * (z + 1.0) ** 3 * self.inv_efunc(z) ** 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import io
import json

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from astropy import table
from astropy import units as u
from astropy.table import MaskedColumn, Table
from astropy.time import Time, TimeDelta
from astropy.utils import minversion
from astropy.utils.compat.optional_deps import (
    HAS_DASK,
    HAS_DUCKDB,
    HAS_NARWHALS,
    HAS_PANDAS,
    HAS_POLARS,
    HAS_PYARROW,
)

from .conftest import MIXIN_COLS


@pytest.mark.parametrize(
    "backend,use_legacy_pandas_api",
    [
        pytest.param(
            "pandas",
            True,
            marks=pytest.mark.skipif(not HAS_PANDAS, reason="requires pandas"),
            id="pandas-legacy",
        ),
        pytest.param(
            "pandas",
            False,
            marks=pytest.mark.skipif(
                not HAS_NARWHALS or not HAS_PANDAS,
                reason="requires narwhals and pandas",
            ),
            id="pandas-generic",
        ),
        pytest.param(
            "polars",
            False,
            marks=pytest.mark.skipif(
                not HAS_NARWHALS or not HAS_POLARS,
                reason="requires narwhals and polars",
            ),
            id="polars-generic",
        ),
        pytest.param(
            "pyarrow",
            False,
            marks=pytest.mark.skipif(
                not HAS_NARWHALS or not HAS_PYARROW,
                reason="requires narwhals and pyarrow",
            ),
            id="pyarrow-generic",
        ),
        pytest.param(
            "dask",
            False,
            marks=pytest.mark.skipif(
                not HAS_NARWHALS or not HAS_DASK,
                reason="requires narwhals and dask",
            ),
            id="dask-generic",
        ),
        pytest.param(
            "duckdb",
            False,
            marks=pytest.mark.skipif(
                not HAS_NARWHALS or not HAS_DUCKDB,
                reason="requires narwhals and duckdb",
            ),
            id="duckdb-generic",
        ),
    ],
)
class TestDataFrameConversion:
    """Test DataFrame conversion functionality for both legacy pandas and generic backends."""

    def _to_dataframe(self, table, backend, use_legacy_pandas_api, **kwargs):
        """Convert table to dataframe using appropriate method."""
        if use_legacy_pandas_api:
            if backend != "pandas":
                raise ValueError(
                    "Legacy conversion is only supported for the pandas backend."
                )
            return table.to_pandas(**kwargs)

        # Lazy backends cannot be exported to
        if backend in ("dask", "duckdb"):
            pytest.skip("Lazy backends cannot be converted back to Table")

        return table.to_df(backend, **kwargs)

    def _from_dataframe(self, df, backend, use_legacy_pandas_api, **kwargs):
        """Convert dataframe to table using appropriate method."""
        if use_legacy_pandas_api:
            if backend != "pandas":
                raise ValueError(
                    "Legacy conversion is only supported for the pandas backend."
                )
            return table.Table.from_pandas(df, **kwargs)
        return table.Table.from_df(df, **kwargs)

    def test_simple(self, backend, use_legacy_pandas_api):
        """Test basic endianness and data type handling."""
        t = table.Table()

        for endian in ["<", ">", "="]:
            for kind in ["f", "i"]:
                for byte in ["2", "4", "8"]:
                    dtype = np.dtype(endian + kind + byte)
                    x = np.array([1, 2, 3], dtype=dtype)
                    t[endian + kind + byte] = x.view(x.dtype.newbyteorder(endian))

        t["u"] = ["a", "b", "c"]
        t["s"] = [b"a", b"b", b"c"]

        d = self._to_dataframe(t, backend, use_legacy_pandas_api)

        # Basic round-trip test
        t2 = self._from_dataframe(d, backend, use_legacy_pandas_api)

        if minversion("pandas", "3.0.0.dev"):
            # upstream feature of pandas
            import pandas as pd

            pandas_string_dtype = pd.StringDtype(na_value=np.nan)
        else:
            # PANDAS_LT_3_0
            pandas_string_dtype = np.dtype("O")

        for column in t.columns:
            original_col = t[column]
            roundtrip_col = t2[column]

            assert roundtrip_col.dtype.kind == original_col.dtype.kind

            if original_col.dtype.kind in ("U", "S"):
                assert_array_equal(roundtrip_col, original_col)
                if original_col.dtype.kind == "U" and backend == "pandas":
                    # Pandas-specific checks
                    assert d[column].dtype == pandas_string_dtype
            else:
                # Generic comparison with tolerance
                assert_allclose(roundtrip_col, original_col)

                # Compare dtypes by value, not identity (normalize endianness)
                t_dtype = original_col.dtype.newbyteorder("=")
                t2_dtype = roundtrip_col.dtype.newbyteorder("=")

                if backend == "polars" and "f2" in column:
                    # No Polars Float16 support
                    pass
                else:
                    assert t2_dtype == t_dtype

                # Pandas-specific checks
                if backend == "pandas":
                    # Pandas-specific exact comparison
                    assert_array_equal(d[column], t[column])
                    if t[column].dtype.isnative:
                        assert d[column].dtype == t[column].dtype
                    else:
                        assert d[column].dtype == t[column].dtype.newbyteorder()

        # Pandas-specific endian tests (skip if not pandas)
        if not backend == "pandas":
            return

        # Regression test for astropy/astropy#1156 - the following code gave a
        # ValueError: Big-endian buffer not supported on little-endian
        # compiler. We now automatically swap the endian-ness to native order
        # upon adding the arrays to the data frame.
        # Explicitly testing little/big/native endian separately -
        # regression for a case in astropy/astropy#11286 not caught by #3729.
        d[["<i4", ">i4"]]
        d[["<f4", ">f4"]]

        # Additional round-trip checks
        for column in t.columns:
            if column in ("u", "s"):
                assert_array_equal(t2[column], t[column])
            else:
                assert_allclose(t2[column], t[column])
            if t[column].dtype.isnative:
                assert t2[column].dtype == t[column].dtype
            else:
                assert t2[column].dtype == t[column].dtype.newbyteorder()

    def test_from_df_simple(self, backend, use_legacy_pandas_api):
        data = {"a": [1, 2, 3], "b": [4.0, 5.0, 6.0], "c": ["x", "y", "z"]}
        match backend:
            case "pandas":
                import pandas as pd

                df = pd.DataFrame(data)
            case "polars":
                import polars as pl

                df = pl.DataFrame(data)
            case "pyarrow":
                import pyarrow as pa

                df = pa.Table.from_pydict(data)
            case "dask":
                import dask.array as da
                import dask.dataframe as dd

                df = dd.concat(
                    [
                        dd.from_dask_array(
                            da.from_array(arr, chunks="auto"), columns=col
                        )
                        for col, arr in data.items()
                    ],
                    axis=1,
                )
            case "duckdb":
                import duckdb

                df = duckdb.read_json(
                    io.StringIO(
                        json.dumps(
                            [
                                dict(zip(data.keys(), values, strict=True))
                                for values in zip(*data.values(), strict=True)
                            ]
                        )
                    )
                )

            case _:
                raise ValueError(f"Unknown backend: {backend}")

        # Test if the columns are as expected
        tb = self._from_dataframe(df, backend, use_legacy_pandas_api)

        for col in tb.colnames:
            assert_array_equal(tb[col], data[col])

    def test_int128(self, backend, use_legacy_pandas_api):
        match backend:
            case "polars":
                import polars as pl

                df = pl.DataFrame(
                    [pl.Series(name="x", values=range(10), dtype=pl.Int128)]
                )

            case "pandas" | "pyarrow" | "dask" | "duckdb":
                return

            case _:
                raise ValueError(f"Unknown backend: {backend}")

        with pytest.raises(
            ValueError,
            match="Astropy Table does not support narwhals.Int128",
        ):
            _ = self._from_dataframe(df, backend, use_legacy_pandas_api)

    @pytest.mark.parametrize("use_IndexedTable", [False, True])
    def test_to_df_index(self, backend, use_legacy_pandas_api, use_IndexedTable):
        """Test indexing options for both legacy pandas and generic backends."""

        class IndexedTable(table.QTable):
            """Always index the first column"""

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.add_index(self.colnames[0])

        tm = Time([1998, 2002], format="jyear")
        x = [1, 2]
        table_cls = IndexedTable if use_IndexedTable else table.QTable
        t = table_cls([tm, x], names=["tm", "x"])

        match backend:
            case "pandas":
                import pandas as pd

                row_index = pd.RangeIndex(0, 2, 1)
                tm_index = pd.DatetimeIndex(
                    ["1998-01-01", "2002-01-01"],
                    dtype="datetime64[ns]",
                    name="tm",
                    freq=None,
                )
                tp = self._to_dataframe(t, backend, use_legacy_pandas_api)

            case "polars" | "pyarrow":
                with pytest.raises(
                    ValueError,
                    match="Indexing is only supported for pandas-like backends",
                ):
                    self._to_dataframe(t, backend, use_legacy_pandas_api, index="tm")
                return
            case "dask" | "duckdb":
                # Lazy backends will raise ValueError in _to_dataframe
                self._to_dataframe(t, backend, use_legacy_pandas_api)
                return
            case _:
                raise ValueError(f"Unknown backend: {backend}")

        if not use_IndexedTable:
            assert_array_equal(tp.index, row_index)
            tp = self._to_dataframe(t, backend, use_legacy_pandas_api, index="tm")
            assert_array_equal(tp.index, tm_index)
            t.add_index("tm")

        tp = self._to_dataframe(t, backend, use_legacy_pandas_api)
        assert_array_equal(tp.index, tm_index)
        # Make sure writing to dataframe didn't hack the original table
        assert t["tm"].info.indices

        tp = self._to_dataframe(t, backend, use_legacy_pandas_api, index=True)
        assert_array_equal(tp.index, tm_index)

        tp = self._to_dataframe(t, backend, use_legacy_pandas_api, index=False)
        assert_array_equal(tp.index, row_index)

        with pytest.raises(ValueError, match="is not in the table columns"):
            self._to_dataframe(t, backend, use_legacy_pandas_api, index="not a column")

    def test_from_df_index(self, backend, use_legacy_pandas_api):
        """Test index handling in from_dataframe conversion."""
        tm = Time([1998, 2002], format="jyear")
        x = [1, 2]
        t = table.Table([tm, x], names=["tm", "x"])
        match backend:
            case "polars" | "pyarrow":
                # Non-pandas backends don't support indexing
                with pytest.raises(
                    ValueError,
                    match="Indexing is only supported for pandas-like backends",
                ):
                    self._to_dataframe(t, backend, use_legacy_pandas_api, index="tm")
                return
            case "dask" | "duckdb":
                # Lazy backends will raise ValueError in _to_dataframe
                self._to_dataframe(t, backend, use_legacy_pandas_api, index="tm")
                return
            case "pandas":
                tp = self._to_dataframe(t, backend, use_legacy_pandas_api, index="tm")
            case _:
                raise ValueError(f"Unknown backend: {backend}")

        t2 = self._from_dataframe(tp, backend, use_legacy_pandas_api)
        assert t2.colnames == ["x"]

        t2 = self._from_dataframe(tp, backend, use_legacy_pandas_api, index=True)
        assert t2.colnames == ["tm", "x"]
        assert np.allclose(t2["tm"].jyear, tm.jyear)

    def test_units(self, backend, use_legacy_pandas_api):
        """Test handling of units in from_dataframe conversion."""
        data = {"x": [1, 2, 3], "t": [1.3, 1.2, 1.8]}
        match backend:
            case "pandas":
                import pandas as pd

                df = pd.DataFrame(data)
            case "polars":
                import polars as pl

                df = pl.DataFrame(data)
            case "pyarrow":
                import pyarrow as pa

                df = pa.Table.from_pydict(data)
            case "dask":
                import dask.array as da
                import dask.dataframe as dd

                # Need to aggregate data into 2D dask array
                df = dd.concat(
                    [
                        dd.from_dask_array(
                            da.from_array(arr, chunks="auto"), columns=col
                        )
                        for col, arr in data.items()
                    ],
                    axis=1,
                )
            case "duckdb":
                import duckdb

                # DuckDB can't read directly from dict, need to stream via JSON
                df = duckdb.read_json(
                    io.StringIO(
                        json.dumps(
                            [
                                dict(zip(data.keys(), values), strict=True)
                                for values in zip(*data.values(), strict=True)
                            ]
                        )
                    )
                )
            case _:
                raise ValueError(f"Unknown backend: {backend}")

        t = self._from_dataframe(
            df, backend, use_legacy_pandas_api, units={"x": u.m, "t": u.s}
        )

        assert t["x"].unit == u.m
        assert t["t"].unit == u.s

        # test error if not a mapping
        with pytest.raises(TypeError):
            self._from_dataframe(df, backend, use_legacy_pandas_api, units=[u.m, u.s])

        # test warning is raised if additional columns in units dict
        with pytest.warns(UserWarning) as record:
            self._from_dataframe(
                df, backend, use_legacy_pandas_api, units={"x": u.m, "t": u.s, "y": u.m}
            )
        assert len(record) == 1
        assert "{'y'}" in str(record[0].message.args[0])

    @pytest.mark.parametrize("unsigned", ["u", ""])
    @pytest.mark.parametrize("bits", [8, 16, 32, 64])
    def test_nullable_int(self, backend, use_legacy_pandas_api, unsigned, bits):
        """Test nullable integer handling."""
        np_dtype = f"{unsigned}int{bits}"
        c = MaskedColumn([1, 2], mask=[False, True], dtype=np_dtype)
        t = Table([c])
        df = self._to_dataframe(t, backend, use_legacy_pandas_api)

        t2 = self._from_dataframe(df, backend, use_legacy_pandas_api)
        assert str(t2["col0"].dtype) == np_dtype
        assert_array_equal(t2["col0"].mask, [False, True])
        assert_array_equal(t2["col0"], c)

    @pytest.mark.parametrize("ndim", [1, 2, 3])
    def test_nd_columns(self, backend, use_legacy_pandas_api, ndim):
        """Test handling of multidimensional columns."""
        # Add one since we want the dimension of each entry to be ndim
        shape = (3,) * ndim
        colshape = (10,) + shape
        t = table.Table()
        t["a"] = np.arange(np.prod(colshape)).reshape(colshape)

        match backend:
            # Pandas and PyArrow do not support multidimensional columns
            case "pandas" | "pyarrow":
                if ndim > 1:
                    with pytest.raises(
                        ValueError,
                        match="Cannot convert a table with multidimensional columns",
                    ):
                        self._to_dataframe(t, backend, use_legacy_pandas_api)
                    return
            case "dask" | "duckdb":
                # Lazy backends will raise ValueError in _to_dataframe
                self._to_dataframe(t, backend, use_legacy_pandas_api)
                return
            case "polars":
                df = self._to_dataframe(t, backend, use_legacy_pandas_api)
                # Convert and check shape
                assert df["a"].dtype.shape == shape

                # Round-trip conversion
                t2 = self._from_dataframe(df, backend, use_legacy_pandas_api)
                assert t2["a"].shape == colshape
                assert_array_equal(t["a"], t2["a"])
            case _:
                raise ValueError(f"Unknown backend: {backend}")

    def test_mixin_columns(self, backend, use_legacy_pandas_api):
        """Test handling of astropy mixin columns."""
        t = table.QTable()
        for name in sorted(MIXIN_COLS):
            if not name.startswith("ndarray"):
                t[name] = MIXIN_COLS[name]

        t["dt"] = TimeDelta([0, 2, 4, 6], format="sec")

        tp = self._to_dataframe(t, backend, use_legacy_pandas_api)
        t2 = self._from_dataframe(tp, backend, use_legacy_pandas_api)

        assert np.allclose(t2["quantity"], [0, 1, 2, 3])
        assert np.allclose(t2["longitude"], [0.0, 1.0, 5.0, 6.0])
        assert np.allclose(t2["latitude"], [5.0, 6.0, 10.0, 11.0])
        assert np.allclose(t2["skycoord.ra"], [0, 1, 2, 3])
        assert np.allclose(t2["skycoord.dec"], [0, 1, 2, 3])
        assert np.allclose(t2["arraywrap"], [0, 1, 2, 3])
        assert np.allclose(t2["arrayswap"], [0, 1, 2, 3])
        assert np.allclose(
            t2["earthlocation.y"], [0, 110708, 547501, 654527], rtol=0, atol=1
        )

        # Time and TimeDelta mixins that round-trip the class
        assert type(t2["time"]) is Time
        assert np.allclose(t2["time"].jyear, [2000, 2001, 2002, 2003])
        assert np.all(
            t2["time"].isot
            == [
                "2000-01-01T12:00:00.000",
                "2000-12-31T18:00:00.000",
                "2002-01-01T00:00:00.000",
                "2003-01-01T06:00:00.000",
            ]
        )
        assert t2["time"].format == "isot"

        # TimeDelta
        assert type(t2["dt"]) is TimeDelta
        assert np.allclose(t2["dt"].value, [0, 2, 4, 6])
        assert t2["dt"].format == "sec"

    def test_mixin_masked(self, backend, use_legacy_pandas_api):
        """Test handling of masked mixin columns."""
        tm = Time([1, 2, 3], format="cxcsec")
        dt = TimeDelta([1, 2, 3], format="sec")
        tm[1] = np.ma.masked
        dt[1] = np.ma.masked
        t = table.QTable([tm, dt], names=["tm", "dt"])

        tp = self._to_dataframe(t, backend, use_legacy_pandas_api)

        match backend:
            case "pandas":
                tm_nulls = tp["tm"].isnull().to_list()
                dt_nulls = tp["dt"].isnull().to_list()
            case "polars":
                tm_nulls = tp["tm"].is_null().to_list()
                dt_nulls = tp["dt"].is_null().to_list()
            case "pyarrow":
                tm_nulls = tp["tm"].is_null().to_pylist()
                dt_nulls = tp["dt"].is_null().to_pylist()
            case _:
                raise ValueError(f"Unknown backend: {backend}")

        # Common assertion for all backends
        expected_nulls = [False, True, False]
        assert tm_nulls == expected_nulls
        assert dt_nulls == expected_nulls

        # Round-trip conversion
        t2 = self._from_dataframe(tp, backend, use_legacy_pandas_api)

        assert_array_equal(t2["tm"].mask, tm.mask)
        assert np.ma.allclose(t2["tm"].jd, tm.jd, rtol=1e-14, atol=1e-14)

        assert_array_equal(t2["dt"].mask, dt.mask)
        assert np.ma.allclose(t2["dt"].jd, dt.jd, rtol=1e-14, atol=1e-14)

    @pytest.mark.parametrize("use_nullable_int", [True, False])
    def test_masking(self, backend, use_legacy_pandas_api, use_nullable_int):
        """Test handling of masked columns."""
        t = table.Table(masked=True)

        t["a"] = [1, 2, 3]
        t["a"].mask = [True, False, True]

        t["b"] = [1.0, 2.0, 3.0]
        t["b"].mask = [False, False, True]

        t["u"] = ["a", "b", "c"]
        t["u"].mask = [False, True, False]

        t["s"] = [b"a", b"b", b"c"]
        t["s"].mask = [False, True, False]

        # https://github.com/astropy/astropy/issues/7741
        t["Source"] = [2584290278794471936, 2584290038276303744, 2584288728310999296]
        t["Source"].mask = [False, False, False]

        if use_nullable_int:  # Default
            df = self._to_dataframe(
                t, backend, use_legacy_pandas_api, use_nullable_int=use_nullable_int
            )
        else:
            with pytest.raises(
                ValueError,
                match="Cannot convert masked integer columns to DataFrame without using nullable integers.",
            ):
                df = self._to_dataframe(
                    t, backend, use_legacy_pandas_api, use_nullable_int=use_nullable_int
                )
            return

        t2 = self._from_dataframe(df, backend, use_legacy_pandas_api)
        for name, column in t.columns.items():
            assert_array_equal(t2[name].data, column.data)
            if hasattr(t2[name], "mask"):
                assert_array_equal(t2[name].mask, column.mask)

            if column.dtype.kind == "i":
                if np.any(column.mask) and not use_nullable_int:
                    assert t2[name].dtype.kind == "f"
                else:
                    assert t2[name].dtype.kind == "i"
            else:
                if column.dtype.byteorder in ("=", "|"):
                    assert t2[name].dtype == column.dtype
                else:
                    assert t2[name].dtype == column.dtype.newbyteorder()

    def test_basic_roundtrip(self, backend, use_legacy_pandas_api):
        """Test basic round-trip conversion for different backends."""
        t = table.Table()
        t["a"] = [1, 2, 3]
        t["b"] = [4.0, 5.0, 6.0]
        t["c"] = ["x", "y", "z"]

        # Convert to DataFrame and back
        df = self._to_dataframe(t, backend, use_legacy_pandas_api)
        t2 = self._from_dataframe(df, backend, use_legacy_pandas_api)

        # Check that data is preserved
        assert_allclose(t2["a"], t["a"])
        assert_allclose(t2["b"], t["b"])
        assert_array_equal(t2["c"], t["c"])

    def test_units_preservation(self, backend, use_legacy_pandas_api):
        """Test that units are handled correctly through DataFrame conversion."""
        t = table.QTable()
        t["x"] = [1, 2, 3] * u.m
        t["y"] = [4.0, 5.0, 6.0] * u.s

        # Test that units are lost in DataFrame conversion (expected behavior)
        df = self._to_dataframe(t, backend, use_legacy_pandas_api)
        t2 = self._from_dataframe(df, backend, use_legacy_pandas_api)

        # Original table should still have units
        assert t["x"].unit == u.m
        assert t["y"].unit == u.s

        # Round-trip table should not have units
        assert t2["x"].unit is None
        assert t2["y"].unit is None

        # But data should be preserved
        assert_allclose(t2["x"], t["x"].value)
        assert_allclose(t2["y"], t["y"].value)

    def test_masked_int_data(self, backend, use_legacy_pandas_api):
        """Test specific masked integer data handling."""
        data = {"data": [0, 1, 2]}
        t = table.Table(data=data, masked=True)
        t["data"].mask = [1, 1, 0]

        df = self._to_dataframe(t, backend, use_legacy_pandas_api)

        match backend:
            case "pandas":
                val = df["data"].iloc[2]
                nulls_first_two = df["data"].isnull().iloc[:2]
            case "polars":
                val = df["data"][2]
                nulls_first_two = df["data"].is_null()[:2]
            case "pyarrow":
                val = df["data"][2].as_py()
                nulls_first_two = df["data"].is_null()[:2].to_numpy()
            case _:
                raise ValueError(f"Unknown backend: {backend}")

        assert val == 2
        assert nulls_first_two.all()


@pytest.mark.skipif(
    not HAS_PANDAS or not HAS_NARWHALS,
    reason="requires pandas and narwhals",
)
@pytest.mark.parametrize("method", ["from_df", "from_pandas"])
def test_from_pandas_df_with_qtable(method):
    """Test fix for QTable.from_pandas / from_df returns Table not QTable #18909"""
    t = table.QTable()
    t["a"] = [1, 2]
    t["q"] = [3.0, 4.0]
    df = t.to_pandas()
    qt = getattr(table.QTable, method)(df)
    assert isinstance(qt, table.QTable)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test numpy functions and ufuncs on Masked arrays and quantities.

The tests here are fairly detailed but do not aim for complete
coverage.  Complete coverage of all numpy functions is done
with less detailed tests in test_function_helpers.
"""

# We generally call the ufunc in the tests, since those can take
# all ufunc arguments (like axes), but also test whether we can
# mask the exceptions and warnings from the wrappers in erfa itself.
import erfa
import erfa.ufunc as erfa_ufunc
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from astropy import units as u
from astropy.units import Quantity
from astropy.utils import minversion
from astropy.utils.compat.numpycompat import NUMPY_LT_1_25
from astropy.utils.masked.core import Masked

from .test_masked import (
    LongitudeSetup,
    MaskedArraySetup,
    QuantitySetup,
    assert_masked_equal,
)


class MaskedUfuncTests(MaskedArraySetup):
    @pytest.mark.parametrize(
        "ufunc", (np.add, np.subtract, np.divide, np.arctan2, np.minimum)
    )
    @pytest.mark.parametrize("a, b", [("ma", "mb"), ("ma", "b"), ("a", "mb")])
    def test_2op_ufunc(self, ufunc, a, b):
        a, b = getattr(self, a), getattr(self, b)
        mask_a = getattr(a, "mask", np.zeros(a.shape, bool))
        mask_b = getattr(b, "mask", np.zeros(b.shape, bool))
        result = ufunc(a, b)
        expected_data = ufunc(self.a, self.b)
        expected_mask = mask_a | mask_b
        # Note: assert_array_equal also checks type, i.e., that, e.g.,
        # Longitude decays into an Angle.
        assert_array_equal(result.unmasked, expected_data)
        assert_array_equal(result.mask, expected_mask)

        out = Masked(np.zeros_like(result.unmasked))
        result2 = ufunc(a, b, out=out)
        assert result2 is out
        assert_masked_equal(result2, result)

    @pytest.mark.parametrize("base_mask", [True, False])
    def test_ufunc_inplace_where(self, base_mask):
        # Construct base filled with -9 and base_mask (copying to get unit/class).
        base = self.ma.copy()
        base.unmasked.view(np.ndarray)[...] = -9.0
        base._mask[...] = base_mask
        out = base.copy()
        where = np.array([[True, False, False], [False, True, False]])
        result = np.add(self.ma, self.mb, out=out, where=where)
        # Direct checks.
        assert np.all(result.unmasked[~where] == base.unmasked[0, 0])
        assert np.all(result.unmasked[where] == (self.a + self.b)[where])
        # Full comparison.
        expected = base.unmasked.copy()
        np.add(self.a, self.b, out=expected, where=where)
        expected_mask = base.mask.copy()
        np.logical_or(self.mask_a, self.mask_b, out=expected_mask, where=where)
        assert_array_equal(result.unmasked, expected)
        assert_array_equal(result.mask, expected_mask)

    @pytest.mark.parametrize("base_mask", [True, False])
    def test_ufunc_inplace_masked_where(self, base_mask):
        base = self.ma.copy()
        base.unmasked.view(np.ndarray)[...] = -9.0
        base._mask[...] = base_mask
        out = base.copy()
        where = Masked(
            [[True, False, True], [False, False, True]],
            mask=[[True, False, False], [True, False, True]],
        )
        result = np.add(self.ma, self.mb, out=out, where=where)
        # Direct checks.
        assert np.all(result.unmasked[~where.unmasked] == base.unmasked[0, 0])
        assert np.all(
            result.unmasked[where.unmasked] == (self.a + self.b)[where.unmasked]
        )
        assert np.all(result.mask[where.mask])
        assert np.all(result.mask[~where.mask & ~where.unmasked] == base.mask[0, 0])
        assert np.all(
            result.mask[~where.mask & where.unmasked]
            == (self.mask_a | self.mask_b)[~where.mask & where.unmasked]
        )
        # Full comparison.
        expected = base.unmasked.copy()
        np.add(self.a, self.b, out=expected, where=where.unmasked)
        expected_mask = base.mask.copy()
        np.logical_or(self.mask_a, self.mask_b, out=expected_mask, where=where.unmasked)
        expected_mask |= where.mask
        assert_array_equal(result.unmasked, expected)
        assert_array_equal(result.mask, expected_mask)

    def test_ufunc_inplace_no_masked_input(self):
        a_b = np.add(self.a, self.b)
        out = Masked(np.zeros_like(a_b))
        result = np.add(self.a, self.b, out=out)
        assert result is out
        assert_array_equal(result.unmasked, a_b)
        assert_array_equal(result.mask, np.zeros(a_b.shape, bool))

    def test_ufunc_inplace_error(self):
        # Output is not masked.
        out = np.zeros(self.ma.shape)
        with pytest.raises(TypeError):
            np.add(self.ma, self.mb, out=out)

    @pytest.mark.xfail(NUMPY_LT_1_25, reason="masked where not supported in numpy<1.25")
    def test_ufunc_inplace_error_masked_where(self):
        # Input and output are not masked, but where is.
        # Note: prior to numpy 1.25, we cannot control this.
        out = self.a.copy()
        with pytest.raises(TypeError):
            np.add(self.a, self.b, out=out, where=Masked(True, mask=True))

    @pytest.mark.parametrize("ufunc", (np.add.outer, np.minimum.outer))
    @pytest.mark.parametrize("a, b", [("ma", "mb"), ("ma", "b"), ("a", "mb")])
    def test_2op_ufunc_outer(self, ufunc, a, b):
        a, b = getattr(self, a), getattr(self, b)
        mask_a = getattr(a, "mask", np.zeros(a.shape, bool))
        mask_b = getattr(b, "mask", np.zeros(b.shape, bool))
        result = ufunc(a, b)
        expected_data = ufunc(self.a, self.b)
        expected_mask = np.logical_or.outer(mask_a, mask_b)
        # Note: assert_array_equal also checks type, i.e., that, e.g.,
        # Longitude decays into an Angle.
        assert_array_equal(result.unmasked, expected_data)
        assert_array_equal(result.mask, expected_mask)

        out = Masked(np.zeros_like(result.unmasked))
        result2 = ufunc(a, b, out=out)
        assert result2 is out
        assert_masked_equal(result2, result)

    @pytest.mark.parametrize("ufunc", (np.add.outer, np.minimum.outer))
    def test_2op_ufunc_outer_no_masked_input(self, ufunc):
        expected_data = ufunc(self.a, self.b)
        out = Masked(np.zeros_like(expected_data), True)
        result = ufunc(self.a, self.b, out=out)
        assert_array_equal(out.unmasked, expected_data)
        assert_array_equal(out.mask, np.zeros(out.shape, dtype=bool))

    def test_3op_ufunc(self):
        ma_mb = np.clip(self.ma, self.b, self.c)
        expected_data = np.clip(self.a, self.b, self.c)
        expected_mask = self.mask_a
        assert_array_equal(ma_mb.unmasked, expected_data)
        assert_array_equal(ma_mb.mask, expected_mask)

    def test_multi_op_ufunc(self):
        mask = [True, False, False]
        iy = Masked([2000, 2001, 2002], mask=mask)
        im = Masked([1, 2, 3], mask=mask)
        idy = Masked([10, 20, 25], mask=mask)
        ihr = Masked([11, 12, 13], mask=[False, False, True])
        imn = np.array([50, 51, 52])
        isc = np.array([12.5, 13.6, 14.7])
        result = erfa_ufunc.dtf2d("utc", iy, im, idy, ihr, imn, isc)
        # Also test scalar
        result0 = erfa_ufunc.dtf2d("utc", iy[0], im[0], idy[0], ihr[0], imn[0], isc[0])
        expected = erfa_ufunc.dtf2d(
            "utc", iy.unmasked, im.unmasked, idy.unmasked, ihr.unmasked, imn, isc
        )
        expected_mask = np.array([True, False, True])
        for res, res0, exp in zip(result, result0, expected):
            assert_array_equal(res.unmasked, exp)
            assert_array_equal(res.mask, expected_mask)
            assert res0.unmasked == exp[0]
            assert res0.mask == expected_mask[0]

    def test_erfa_pdp_with_out(self):
        p = np.arange(9.0).reshape(3, 3)
        mp = Masked(p)
        mp.mask[1, 2] = True
        out = Masked(np.empty(3), mask=True)
        result = erfa_ufunc.pdp(mp, mp, out=out)
        assert result is out
        assert_array_equal(result.unmasked, erfa_ufunc.pdp(p, p))
        assert_array_equal(result.mask, [False, True, False])
        # With axes just for the inputs.
        axes = [0, 1]
        result2 = erfa_ufunc.pdp(mp, mp, out=out, axes=axes)
        assert result2 is out
        assert_array_equal(result2.unmasked, erfa_ufunc.pdp(p, p, axes=axes))
        assert_array_equal(result2.mask, [False, True, True])

    @pytest.mark.parametrize("kwargs", [{}, dict(axis=0), dict(axes=[0])])
    def test_erfa_p2s_with_out(self, kwargs):
        p = np.arange(9.0).reshape(3, 3)
        mp = Masked(p)
        mp.mask[1, 2] = True
        # Outputs are theta, phi, r.
        outs = tuple(Masked(np.empty(3), mask=True) for _ in range(3))
        masks = tuple(out.mask for out in outs)
        results = erfa_ufunc.p2s(mp, out=outs, **kwargs)
        assert len(results) == 3
        expected = erfa_ufunc.p2s(mp.unmasked, **kwargs)
        expected_mask = mp.mask.any(0 if kwargs else -1)
        for a, b, m, x in zip(results, outs, masks, expected):
            assert a is b
            assert a.mask is m
            assert_array_equal(a.unmasked, x)
            assert_array_equal(a.mask, expected_mask)

    def test_erfa_rxp(self):
        # Regression tests for gh-16116
        m = Masked(np.eye(3))
        v = Masked(np.arange(6).reshape(2, 3))
        rxp1 = erfa_ufunc.rxp(m, v)
        exp = erfa_ufunc.rxp(m.unmasked, v.unmasked)
        assert_array_equal(rxp1.unmasked, exp)
        assert_array_equal(rxp1.mask, False)
        v.mask[0, 0] = True
        rxp2 = erfa_ufunc.rxp(m, v)
        assert_array_equal(rxp2.unmasked, exp)
        assert_array_equal(rxp2.mask, [[True] * 3, [False] * 3])
        m.mask[1, 1] = True
        v.mask[...] = False
        rxp3 = erfa_ufunc.rxp(m, v)
        assert_array_equal(rxp3.unmasked, exp)
        assert_array_equal(rxp3.mask, True)

    def test_erfa_rxr_axes(self):
        m1 = Masked(np.arange(27.0).reshape(3, 3, 3))
        m2 = Masked(np.arange(-27.0, 0.0).reshape(3, 3, 3))
        rxr1 = erfa_ufunc.rxr(m1, m2)
        exp = erfa_ufunc.rxr(m1.unmasked, m2.unmasked)
        assert_array_equal(rxr1.unmasked, exp)
        assert_array_equal(rxr1.mask, False)
        m1.mask[0, 1, 2] = True
        rxr2 = erfa_ufunc.rxr(m1, m2)
        assert_array_equal(rxr2.unmasked, exp)
        assert np.all(rxr2.mask == [[[True]], [[False]], [[False]]])
        axes = [(0, 2), (-2, -1), (0, 1)]
        rxr3 = erfa_ufunc.rxr(m1, m2, axes=axes)
        exp3 = erfa_ufunc.rxr(m1.unmasked, m2.unmasked, axes=axes)
        assert_array_equal(rxr3.unmasked, exp3)
        assert np.all(rxr3.mask == [False, True, False])

    @pytest.mark.parametrize("axis", (0, 1, None))
    def test_add_reduce(self, axis):
        ma_reduce = np.add.reduce(self.ma, axis=axis)
        expected_data = np.add.reduce(self.a, axis=axis)
        expected_mask = np.logical_or.reduce(self.ma.mask, axis=axis)
        assert_array_equal(ma_reduce.unmasked, expected_data)
        assert_array_equal(ma_reduce.mask, expected_mask)

        out = Masked(np.zeros_like(ma_reduce.unmasked), np.ones_like(ma_reduce.mask))
        ma_reduce2 = np.add.reduce(self.ma, axis=axis, out=out)
        assert ma_reduce2 is out
        assert_masked_equal(ma_reduce2, ma_reduce)

    def test_add_reduce_no_masked_input(self):
        a_reduce = np.add.reduce(self.a, axis=0)
        out = Masked(np.zeros_like(a_reduce), np.ones(a_reduce.shape, bool))
        result = np.add.reduce(self.a, axis=0, out=out)
        assert result is out
        assert_array_equal(out.unmasked, a_reduce)
        assert_array_equal(out.mask, False)
        # Also try with where (which should have different path)
        where = np.array([[True, False, False], [True, True, False]])
        a_reduce2 = np.add.reduce(self.a, axis=0, where=where)
        result2 = np.add.reduce(self.a, axis=0, out=out, where=where)
        assert result2 is out
        assert_array_equal(out.unmasked, a_reduce2)
        assert_array_equal(out.mask, [False, False, True])

    @pytest.mark.parametrize("axis", (0, 1, None))
    def test_minimum_reduce(self, axis):
        ma_reduce = np.minimum.reduce(self.ma, axis=axis)
        expected_data = np.minimum.reduce(self.a, axis=axis)
        expected_mask = np.logical_or.reduce(self.ma.mask, axis=axis)
        assert_array_equal(ma_reduce.unmasked, expected_data)
        assert_array_equal(ma_reduce.mask, expected_mask)

    @pytest.mark.parametrize("axis", (0, 1, None))
    def test_maximum_reduce(self, axis):
        ma_reduce = np.maximum.reduce(self.ma, axis=axis)
        expected_data = np.maximum.reduce(self.a, axis=axis)
        expected_mask = np.logical_or.reduce(self.ma.mask, axis=axis)
        assert_array_equal(ma_reduce.unmasked, expected_data)
        assert_array_equal(ma_reduce.mask, expected_mask)


class TestMaskedArrayUfuncs(MaskedUfuncTests):
    # multiply.reduce does not work with units, so test only for plain array.
    # Similarly, modf only works for dimensionless, but we are using it just
    # to check multiple outputs get the right mask.
    @pytest.mark.parametrize("axis", (0, 1, None))
    def test_multiply_reduce(self, axis):
        ma_reduce = np.multiply.reduce(self.ma, axis=axis)
        expected_data = np.multiply.reduce(self.a, axis=axis)
        expected_mask = np.logical_or.reduce(self.ma.mask, axis=axis)
        assert_array_equal(ma_reduce.unmasked, expected_data)
        assert_array_equal(ma_reduce.mask, expected_mask)

    def test_ufunc_two_out(self):
        out0 = np.empty_like(self.ma)
        out0_mask = out0.mask
        out1 = np.empty_like(self.ma)
        out1_mask = out1.mask
        res0, res1 = np.modf(self.ma, out=(out0, out1))
        assert res0 is out0
        assert res1 is out1
        assert out0.mask is out0_mask
        assert out1.mask is out1_mask
        assert_array_equal(out0.mask, self.mask_a)
        assert_array_equal(out1.mask, self.mask_a)

    def test_ufunc_not_implemented_for_other(self):
        """
        If the unmasked operation returns NotImplemented, this
        should lead to a TypeError also for the masked version.
        """
        a = np.array([1, 2])
        b = 3 * u.m
        with pytest.raises(TypeError):
            a & b

        ma = Masked(a)
        with pytest.raises(TypeError):
            ma & b


class TestMaskedQuantityUfuncs(MaskedUfuncTests, QuantitySetup):
    def test_ufunc_inplace_error2(self):
        out = Masked(np.zeros(self.ma.shape))
        with pytest.raises(TypeError):
            np.add(self.ma, self.mb, out=out)


class TestMaskedLongitudeUfuncs(MaskedUfuncTests, LongitudeSetup):
    def test_ufunc_inplace_quantity_initial(self):
        out = Masked(np.zeros(self.ma.shape) << u.m)
        result = np.add(self.ma, self.mb, out=out)
        assert result is out
        expected = np.add(self.ma, self.mb).view(Quantity)
        assert_masked_equal(result, expected)


class TestMaskedArrayConcatenation(MaskedArraySetup):
    def test_concatenate(self):
        mb = self.mb[np.newaxis]
        concat_a_b = np.concatenate((self.ma, mb), axis=0)
        expected_data = np.concatenate((self.a, self.b[np.newaxis]), axis=0)
        expected_mask = np.concatenate((self.mask_a, self.mask_b[np.newaxis]), axis=0)
        assert_array_equal(concat_a_b.unmasked, expected_data)
        assert_array_equal(concat_a_b.mask, expected_mask)

    def test_concatenate_not_all_masked(self):
        mb = self.mb[np.newaxis]
        concat_a_b = np.concatenate((self.a, mb), axis=0)
        expected_data = np.concatenate((self.a, self.b[np.newaxis]), axis=0)
        expected_mask = np.concatenate(
            (np.zeros(self.a.shape, bool), self.mask_b[np.newaxis]), axis=0
        )
        assert_array_equal(concat_a_b.unmasked, expected_data)
        assert_array_equal(concat_a_b.mask, expected_mask)

    @pytest.mark.parametrize("obj", (1, slice(2, 3)))
    def test_insert(self, obj):
        mc_in_a = np.insert(self.ma, obj, self.mc, axis=-1)
        expected = Masked(
            np.insert(self.a, obj, self.c, axis=-1),
            np.insert(self.mask_a, obj, self.mask_c, axis=-1),
        )
        assert_masked_equal(mc_in_a, expected)

    def test_insert_masked_obj(self):
        with pytest.raises(TypeError):
            np.insert(self.ma, Masked(1, mask=False), self.mc, axis=-1)

    def test_append(self):
        mc_to_a = np.append(self.ma, self.mc, axis=-1)
        expected = Masked(
            np.append(self.a, self.c, axis=-1),
            np.append(self.mask_a, self.mask_c, axis=-1),
        )
        assert_masked_equal(mc_to_a, expected)


class TestMaskedQuantityConcatenation(TestMaskedArrayConcatenation, QuantitySetup):
    pass


class TestMaskedLongitudeConcatenation(TestMaskedArrayConcatenation, LongitudeSetup):
    pass


class TestMaskedArrayBroadcast(MaskedArraySetup):
    def test_broadcast_to(self):
        shape = self.ma.shape
        ba = np.broadcast_to(self.mb, shape, subok=True)
        assert ba.shape == shape
        assert ba.mask.shape == shape
        expected = Masked(
            np.broadcast_to(self.mb.unmasked, shape, subok=True),
            np.broadcast_to(self.mb.mask, shape, subok=True),
        )
        assert_masked_equal(ba, expected)

    def test_broadcast_to_using_apply(self):
        # Partially just to ensure we cover the relevant part of _apply.
        shape = self.ma.shape
        ba = self.mb._apply(np.broadcast_to, shape=shape, subok=True)
        assert ba.shape == shape
        assert ba.mask.shape == shape
        expected = Masked(
            np.broadcast_to(self.mb.unmasked, shape, subok=True),
            np.broadcast_to(self.mb.mask, shape, subok=True),
        )
        assert_masked_equal(ba, expected)

    def test_broadcast_arrays(self):
        mb = np.broadcast_arrays(self.ma, self.mb, self.mc, subok=True)
        b = np.broadcast_arrays(self.a, self.b, self.c, subok=True)
        bm = np.broadcast_arrays(self.mask_a, self.mask_b, self.mask_c)
        for mb_, b_, bm_ in zip(mb, b, bm):
            assert_array_equal(mb_.unmasked, b_)
            assert_array_equal(mb_.mask, bm_)

    def test_broadcast_arrays_not_all_masked(self):
        mb = np.broadcast_arrays(self.a, self.mb, self.c, subok=True)
        assert_array_equal(mb[0], self.a)
        expected1 = np.broadcast_to(self.mb, self.a.shape, subok=True)
        assert_masked_equal(mb[1], expected1)
        expected2 = np.broadcast_to(self.c, self.a.shape, subok=True)
        assert_array_equal(mb[2], expected2)

    def test_broadcast_arrays_subok_false(self):
        # subok affects ndarray subclasses but not masking itself.
        mb = np.broadcast_arrays(self.ma, self.mb, self.mc, subok=False)
        assert all(type(mb_.unmasked) is np.ndarray for mb_ in mb)
        b = np.broadcast_arrays(self.a, self.b, self.c, subok=False)
        mask_b = np.broadcast_arrays(self.mask_a, self.mask_b, self.mask_c, subok=False)
        for mb_, b_, mask_ in zip(mb, b, mask_b):
            assert_array_equal(mb_.unmasked, b_)
            assert_array_equal(mb_.mask, mask_)


class TestMaskedQuantityBroadcast(TestMaskedArrayBroadcast, QuantitySetup):
    pass


class TestMaskedLongitudeBroadcast(TestMaskedArrayBroadcast, LongitudeSetup):
    pass


class TestMaskedArrayCalculation(MaskedArraySetup):
    @pytest.mark.parametrize("n,axis", [(1, -1), (2, -1), (1, 0)])
    def test_diff(self, n, axis):
        mda = np.diff(self.ma, n=n, axis=axis)
        expected_data = np.diff(self.a, n, axis)
        nan_mask = np.zeros_like(self.a)
        nan_mask[self.ma.mask] = np.nan
        expected_mask = np.isnan(np.diff(nan_mask, n=n, axis=axis))
        assert_array_equal(mda.unmasked, expected_data)
        assert_array_equal(mda.mask, expected_mask)

    def test_diff_explicit(self):
        ma = Masked(
            np.arange(8.0), [True, False, False, False, False, True, False, False]
        )
        mda = np.diff(ma)
        assert np.all(mda.unmasked == 1.0)
        assert np.all(mda.mask == [True, False, False, False, True, True, False])
        mda = np.diff(ma, n=2)
        assert np.all(mda.unmasked == 0.0)
        assert np.all(mda.mask == [True, False, False, True, True, True])


class TestMaskedQuantityCalculation(TestMaskedArrayCalculation, QuantitySetup):
    pass


class TestMaskedLongitudeCalculation(TestMaskedArrayCalculation, LongitudeSetup):
    pass


class TestMaskedArraySorting(MaskedArraySetup):
    @pytest.mark.parametrize("axis", [-1, 0])
    def test_lexsort1(self, axis):
        ma_lexsort = np.lexsort((self.ma,), axis=axis)
        filled = self.a.copy()
        filled[self.mask_a] = 9e9
        expected_data = filled.argsort(axis)
        assert_array_equal(ma_lexsort, expected_data)

    @pytest.mark.parametrize("axis", [-1, 0])
    def test_lexsort2(self, axis):
        mb = np.broadcast_to(-self.mb, self.ma.shape).copy()
        mamb_lexsort = np.lexsort((self.ma, mb), axis=axis)
        filled_a = self.ma.filled(9e9)
        filled_b = mb.filled(9e9)
        expected_ab = np.lexsort((filled_a, filled_b), axis=axis)
        assert_array_equal(mamb_lexsort, expected_ab)
        mbma_lexsort = np.lexsort((mb, self.ma), axis=axis)
        expected_ba = np.lexsort((filled_b, filled_a), axis=axis)
        assert_array_equal(mbma_lexsort, expected_ba)
        mbma_lexsort2 = np.lexsort(np.stack([mb, self.ma], axis=0), axis=axis)
        assert_array_equal(mbma_lexsort2, expected_ba)

    @pytest.mark.parametrize("axis", [-1, 0])
    def test_lexsort_mix(self, axis):
        mb = np.broadcast_to(-self.mb, self.ma.shape).copy()
        mamb_lexsort = np.lexsort((self.a, mb), axis=axis)
        filled_b = mb.filled(9e9)
        expected_ab = np.lexsort((self.a, filled_b), axis=axis)
        assert_array_equal(mamb_lexsort, expected_ab)
        mbma_lexsort = np.lexsort((mb, self.a), axis=axis)
        expected_ba = np.lexsort((filled_b, self.a), axis=axis)
        assert_array_equal(mbma_lexsort, expected_ba)
        mbma_lexsort2 = np.lexsort(np.stack([mb, self.a], axis=0), axis=axis)
        assert_array_equal(mbma_lexsort2, expected_ba)


class TestStructuredUfuncs:
    """Test with structure dtypes, using erfa ufuncs."""

    def test_erfa_d2tf_tf2d(self):
        mask = np.array([True, False, False])
        days = Masked([0.25, 0.875, 0.0625], mask=mask)
        sign, ihmsf = erfa_ufunc.d2tf(3, days)
        assert_array_equal(sign.mask["sign"], mask)
        sign = sign.view("S1")  # Like is done by the erfa wrapper.
        assert_array_equal(sign.mask, mask)
        for name in ihmsf.dtype.names:
            assert_array_equal(ihmsf[name].mask, mask)

        # Check roundtrip.
        check, stat = erfa_ufunc.tf2d(
            sign, ihmsf["h"], ihmsf["m"], ihmsf["s"] + ihmsf["f"]
        )
        assert_allclose(check.unmasked, days, atol=1e-3 / 24 / 3600)
        assert_array_equal(check.mask, mask)
        assert_array_equal(stat.unmasked, 0)
        assert_array_equal(stat.mask, mask)

    def test_erfa_astrom(self):
        mask = np.array([True, False, False])
        jd2 = Masked([0, 0.401182685, 0.5], mask=mask)
        astrom, eo = erfa_ufunc.apci13(2456165.5, jd2)
        assert_array_equal(eo.mask, mask)
        for n in astrom.dtype.names:
            # .T for multi-element fields.
            assert np.all(astrom[n].mask.T == mask)

        along = np.array([0.125, 0.25, 0.35])
        # Not going to worry about different masks for different elements.
        # In principle aper could propagate just what it needs.
        astrom["along"] = Masked([0.125, 0.25, 0.35], mask)
        astrom2 = erfa_ufunc.aper(Masked(np.ones(3), [False, True, False]), astrom)
        assert_array_equal(astrom2["eral"].unmasked, along + 1.0)
        mask2 = mask | [False, True, False]
        for n in astrom2.dtype.names:
            # .T for multi-element fields.
            assert np.all(astrom2[n].mask.T == mask2)

    def test_erfa_atioq(self):
        # Regression test for gh-16123, using test from erfa.
        astrom, _ = erfa_ufunc.apio13(
            2456384.5,
            0.969254051,
            0.1550675,
            -0.527800806,
            -1.2345856,
            2738.0,
            2.47230737e-7,
            1.82640464e-6,
            731.0,
            12.8,
            0.59,
            0.55,
        )
        astrom = Masked(astrom)
        ri = 2.710121572969038991
        di = 0.1729371367218230438
        aob, zob, hob, dob, rob = erfa_ufunc.atioq(ri, di, astrom)
        assert isinstance(aob, Masked)
        # Really should not need to check the values, since done
        # in units/tests/test_quantity_erfa_ufuncs, but why not...
        assert_allclose(aob, 0.9233952224895122499e-1, atol=1e-12, rtol=0)
        assert_allclose(zob, 1.407758704513549991, atol=1e-12, rtol=0)
        assert_allclose(hob, -0.9247619879881698140e-1, atol=1e-12, rtol=0)
        assert_allclose(dob, 0.1717653435756234676, atol=1e-12, rtol=0)
        assert_allclose(rob, 2.710085107988480746, atol=1e-12, rtol=0)


def test_erfa_no_warnings_on_masked_entries():
    # Erfa warns for invalid inputs for some routines.
    msg = 'ERFA function "tf2d" yielded {count} of "ihour outside range 0-23"'
    ihour1 = [25, 26, 10]
    with pytest.warns(erfa.ErfaWarning, match=msg.format(count=2)):
        res1 = erfa.tf2d("+", ihour1, 0, 0.0)
    # But will not if they are masked.
    mask = [True, True, False]
    ihour2 = Masked(ihour1, mask)
    res2 = erfa.tf2d("+", ihour2, 0, 0.0)
    assert_array_equal(res2.unmasked, res1)
    assert_array_equal(res2.mask, mask)
    # And will count correctly.
    mask = [True, False, False]
    ihour3 = Masked(ihour1, mask)
    count = 1 if minversion(erfa, "2.0.1.1", inclusive=False) else "—"
    with pytest.warns(erfa.ErfaWarning, match=msg.format(count=count)):
        res3 = erfa.tf2d("+", ihour3, 0, 0.0)
    assert_array_equal(res3.unmasked, res1)
    assert_array_equal(res3.mask, mask)


def test_erfa_no_exceptions_on_masked_entries():
    # Erfa raises exceptions for invalid inputs in some routines.
    iday1 = [25, 30]
    with pytest.raises(erfa.ErfaError, match="bad day"):
        erfa.dat(2000, 2, iday1, 0.0)
    # But will not if they are masked.
    mask = [False, True]
    iday2 = Masked(iday1, mask)
    res = erfa.dat(2000, 2, iday2, 0.0)
    exp1 = erfa.dat(2000, 2, iday1[0], 0.0)
    assert_array_equal(res.unmasked[0], exp1)
    assert_array_equal(res.mask, mask)



import pickle

import numpy as np
import pytest

import astropy.units as u
from astropy import coordinates as coord
from astropy.coordinates import (
    ICRS,
    Angle,
    Distance,
    DynamicMatrixTransform,
    Latitude,
    Longitude,
    StaticMatrixTransform,
)
from astropy.tests.helper import check_pickling_recovery, pickle_protocol  # noqa: F401


def test_basic():
    lon1 = Longitude(1.23, "radian", wrap_angle="180d")
    s = pickle.dumps(lon1)
    lon2 = pickle.loads(s)


def test_pickle_longitude_wrap_angle():
    a = Longitude(1.23, "radian", wrap_angle="180d")
    s = pickle.dumps(a)
    b = pickle.loads(s)

    assert a.rad == b.rad
    assert a.wrap_angle == b.wrap_angle


def _dummy_transform(fromcoord, toframe):
    return np.identity(3)


@pytest.mark.parametrize(
    "original",
    [
        Angle(0.0 * u.rad),
        Distance(5 * u.pc),
        DynamicMatrixTransform(_dummy_transform, ICRS, ICRS),
        ICRS(0 * u.rad, 0 * u.rad),
        Latitude(0 * u.rad),
        Longitude(0 * u.rad),
        StaticMatrixTransform(np.identity(3), ICRS, ICRS),
    ],
    ids=lambda x: type(x).__name__,
)
def test_simple_object(pickle_protocol, original):  # noqa: F811
    check_pickling_recovery(original, pickle_protocol)


class _CustomICRS(coord.ICRS):
    default_representation = coord.PhysicsSphericalRepresentation


@pytest.mark.parametrize(
    "frame",
    [
        coord.SkyOffsetFrame(origin=coord.ICRS(0 * u.deg, 0 * u.deg)),
        coord.SkyOffsetFrame(
            5 * u.deg, 10 * u.deg, origin=coord.Galactic(2 * u.deg, -3 * u.deg)
        ),
        coord.SkyOffsetFrame(
            5 * u.deg,
            10 * u.deg,
            10 * u.pc,
            origin=coord.Galactic(2 * u.deg, -3 * u.deg),
            representation_type=coord.PhysicsSphericalRepresentation,
        ),
        coord.SkyOffsetFrame(
            5 * u.deg,
            10 * u.deg,
            0 * u.pc,
            origin=_CustomICRS(2 * u.deg, 3 * u.deg, 1 * u.pc),
        ),
    ],
)
def test_skyoffset_pickle(pickle_protocol, frame):  # noqa: F811
    """
    This is a regression test for issue #9249:
    https://github.com/astropy/astropy/issues/9249
    """
    check_pickling_recovery(frame, pickle_protocol)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Defines constants used in `astropy.samp`.
"""

from astropy.utils.data import get_pkg_data_filename

__all__ = [
    "SAFE_MTYPES",
    "SAMP_ICON",
    "SAMP_STATUS_ERROR",
    "SAMP_STATUS_OK",
    "SAMP_STATUS_WARNING",
]

__profile_version__ = "1.3"

#: General constant for samp.ok status string
SAMP_STATUS_OK = "samp.ok"
#: General constant for samp.warning status string
SAMP_STATUS_WARNING = "samp.warning"
#: General constant for samp.error status string
SAMP_STATUS_ERROR = "samp.error"

SAFE_MTYPES = [
    "samp.app.*",
    "samp.msg.progress",
    "table.*",
    "image.*",
    "coord.*",
    "spectrum.*",
    "bibcode.*",
    "voresource.*",
]

with open(get_pkg_data_filename("data/astropy_icon.png"), "rb") as f:
    SAMP_ICON = f.read()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest

import astropy.units as u
from astropy.cosmology._src.utils import (
    aszarr,
    deprecated_keywords,
    vectorize_redshift_method,
)
from astropy.utils.compat.optional_deps import HAS_PANDAS

from .test_core import invalid_zs, valid_zs, z_arr


def test_vectorize_redshift_method():
    """Test :func:`astropy.cosmology._src.utils.vectorize_redshift_method`."""

    class Class:
        @vectorize_redshift_method
        def method(self, z):
            return z

    c = Class()

    assert hasattr(c.method, "__vectorized__")
    assert isinstance(c.method.__vectorized__, np.vectorize)

    # calling with Number
    assert c.method(1) == 1
    assert isinstance(c.method(1), int)

    # calling with a numpy scalar
    assert c.method(np.float64(1)) == np.float64(1)
    assert isinstance(c.method(np.float64(1)), np.float64)

    # numpy array
    assert all(c.method(np.array([1, 2])) == np.array([1, 2]))
    assert isinstance(c.method(np.array([1, 2])), np.ndarray)

    # non-scalar
    assert all(c.method([1, 2]) == np.array([1, 2]))
    assert isinstance(c.method([1, 2]), np.ndarray)


# -------------------------------------------------------------------


class Test_aszarr:
    @pytest.mark.parametrize(
        "z, expect",
        list(
            zip(
                valid_zs,
                [0, 1, 1100, np.float64(3300), 2.0, 3.0, z_arr, z_arr, z_arr, z_arr],
            )
        ),
    )
    def test_valid(self, z, expect):
        """Test :func:`astropy.cosmology._src.utils.aszarr`."""
        got = aszarr(z)
        assert np.array_equal(got, expect)

    @pytest.mark.parametrize("z, exc", invalid_zs)
    def test_invalid(self, z, exc):
        """Test :func:`astropy.cosmology._src.utils.aszarr`."""
        with pytest.raises(exc):
            aszarr(z)

    @pytest.mark.skipif(not HAS_PANDAS, reason="requires pandas")
    def test_pandas(self):
        import pandas as pd

        x = pd.Series([1, 2, 3, 4, 5])

        # Demonstrate Pandas doesn't work with units
        assert not isinstance(x * u.km, u.Quantity)

        # Test aszarr works with Pandas
        assert isinstance(aszarr(x), np.ndarray)
        np.testing.assert_array_equal(aszarr(x), x.values)


# -------------------------------------------------------------------


class TestDeprecatedKeywords:
    @classmethod
    def setup_class(cls):
        def noop(a, b, c, d):
            # a minimal function that does nothing,
            # with multiple positional-or-keywords arguments
            return

        cls.base_func = noop
        cls.depr_funcs = {
            1: deprecated_keywords("a", since="999.999.999")(noop),
            2: deprecated_keywords("a", "b", since="999.999.999")(noop),
            4: deprecated_keywords("a", "b", "c", "d", since="999.999.999")(noop),
        }

    def test_type_safety(self):
        dec = deprecated_keywords(b"a", since="999.999.999")
        with pytest.raises(TypeError, match=r"names\[0\] must be a string"):
            dec(self.base_func)

        dec = deprecated_keywords("a", since=b"999.999.999")
        with pytest.raises(TypeError, match=r"since must be a string"):
            dec(self.base_func)

    @pytest.mark.parametrize("n_deprecated_keywords", [1, 2, 4])
    def test_no_warn(self, n_deprecated_keywords):
        func = self.depr_funcs[n_deprecated_keywords]
        func(1, 2, 3, 4)

    @pytest.mark.parametrize(
        "n_deprecated_keywords, args, kwargs, match",
        [
            pytest.param(
                1,
                (),
                {"a": 1, "b": 2, "c": 3, "d": 4},
                r"Passing 'a' as keyword is deprecated since",
                id="1 deprecation, 1 warn",
            ),
            pytest.param(
                2,
                (1,),
                {"b": 2, "c": 3, "d": 4},
                r"Passing 'b' as keyword is deprecated since",
                id="2 deprecation, 1 warn",
            ),
            pytest.param(
                2,
                (),
                {"a": 1, "b": 2, "c": 3, "d": 4},
                r"Passing \['a', 'b'\] arguments as keywords is deprecated since",
                id="2 deprecations, 2 warns",
            ),
            pytest.param(
                4,
                (),
                {"a": 1, "b": 2, "c": 3, "d": 4},
                (
                    r"Passing \['a', 'b', 'c', 'd'\] arguments as keywords "
                    "is deprecated since"
                ),
                id="4 deprecations, 4 warns",
            ),
            pytest.param(
                4,
                (1,),
                {"b": 2, "c": 3, "d": 4},
                r"Passing \['b', 'c', 'd'\] arguments as keywords is deprecated since",
                id="4 deprecations, 3 warns",
            ),
            pytest.param(
                4,
                (1, 2),
                {"c": 3, "d": 4},
                r"Passing \['c', 'd'\] arguments as keywords is deprecated since",
                id="4 deprecations, 2 warns",
            ),
            pytest.param(
                4,
                (1, 2, 3),
                {"d": 4},
                r"Passing 'd' as keyword is deprecated since",
                id="4 deprecations, 1 warn",
            ),
        ],
    )
    def test_warn(self, n_deprecated_keywords, args, kwargs, match):
        func = self.depr_funcs[n_deprecated_keywords]
        with pytest.warns(FutureWarning, match=match):
            func(*args, **kwargs)



# NOTE: this hook should be added to
# https://github.com/pyinstaller/pyinstaller-hooks-contrib
# once that repository is ready for pull requests
from PyInstaller.utils.hooks import collect_data_files

datas = collect_data_files("skyfield")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module contains functions/values used repeatedly in different modules of
the ``builtin_frames`` package.
"""

import warnings

import erfa
import numpy as np

from astropy import units as u
from astropy.coordinates.earth import EarthLocation
from astropy.coordinates.representation import CartesianDifferential
from astropy.time import Time
from astropy.utils import iers
from astropy.utils.exceptions import AstropyWarning

# We use tt as the time scale for this equinoxes, primarily because it is the
# convention for J2000 (it is unclear if there is any "right answer" for B1950)
# while #8600 makes this the default behavior, we show it here to ensure it's
# clear which is used here
EQUINOX_J2000 = Time("J2000", scale="tt")
EQUINOX_B1950 = Time("B1950", scale="tt")

# This is a time object that is the default "obstime" when such an attribute is
# necessary.  Currently, we use J2000.
DEFAULT_OBSTIME = Time("J2000", scale="tt")

# This is an EarthLocation that is the default "location" when such an attribute is
# necessary. It is the centre of the Earth.
EARTH_CENTER = EarthLocation(0 * u.km, 0 * u.km, 0 * u.km)

PIOVER2 = np.pi / 2.0

# comes from the mean of the 1962-2014 IERS B data
_DEFAULT_PM = (0.035, 0.29) * u.arcsec


def get_polar_motion(time):
    """
    gets the two polar motion components in radians for use with apio.
    """
    # Get the polar motion from the IERS table
    iers_table = iers.earth_orientation_table.get()
    xp, yp, status = iers_table.pm_xy(time, return_status=True)

    wmsg = (
        "Tried to get polar motions for times {} IERS data is "
        "valid. Defaulting to polar motion from the 50-yr mean for those. "
        "This may affect precision at the arcsec level. Please check your "
        "astropy.utils.iers.conf.iers_auto_url and point it to a newer "
        "version if necessary."
    )
    if np.any(status == iers.TIME_BEFORE_IERS_RANGE):
        xp[status == iers.TIME_BEFORE_IERS_RANGE] = _DEFAULT_PM[0]
        yp[status == iers.TIME_BEFORE_IERS_RANGE] = _DEFAULT_PM[1]

        warnings.warn(wmsg.format("before"), AstropyWarning)

    if np.any(status == iers.TIME_BEYOND_IERS_RANGE):
        xp[status == iers.TIME_BEYOND_IERS_RANGE] = _DEFAULT_PM[0]
        yp[status == iers.TIME_BEYOND_IERS_RANGE] = _DEFAULT_PM[1]

        warnings.warn(wmsg.format("after"), AstropyWarning)

    return xp.to_value(u.radian), yp.to_value(u.radian)


def _warn_iers(ierserr):
    """
    Generate a warning for an IERSRangeerror.

    Parameters
    ----------
    ierserr : An `~astropy.utils.iers.IERSRangeError`
    """
    msg = "{0} Assuming UT1-UTC=0 for coordinate transformations."
    warnings.warn(msg.format(ierserr.args[0]), AstropyWarning)


def get_dut1utc(time):
    """
    This function is used to get UT1-UTC in coordinates because normally it
    gives an error outside the IERS range, but in coordinates we want to allow
    it to go through but with a warning.
    """
    try:
        return time.delta_ut1_utc
    except iers.IERSRangeError as e:
        _warn_iers(e)
        return np.zeros(time.shape)


def get_jd12(time, scale):
    """
    Gets ``jd1`` and ``jd2`` from a time object in a particular scale.

    Parameters
    ----------
    time : `~astropy.time.Time`
        The time to get the jds for
    scale : str
        The time scale to get the jds for

    Returns
    -------
    jd1 : float
    jd2 : float
    """
    if time.scale == scale:
        newtime = time
    else:
        try:
            newtime = getattr(time, scale)
        except iers.IERSRangeError as e:
            _warn_iers(e)
            newtime = time

    return newtime.jd1, newtime.jd2


def norm(p):
    """
    Normalise a p-vector.
    """
    return p / np.sqrt(np.einsum("...i,...i", p, p))[..., np.newaxis]


def pav2pv(p, v):
    """
    Combine p- and v- vectors into a pv-vector.
    """
    pv = np.empty(np.broadcast(p, v).shape[:-1], erfa.dt_pv)
    pv["p"] = p
    pv["v"] = v
    return pv


def get_cip(jd1, jd2):
    """
    Find the X, Y coordinates of the CIP and the CIO locator, s.

    Parameters
    ----------
    jd1 : float or `np.ndarray`
        First part of two part Julian date (TDB)
    jd2 : float or `np.ndarray`
        Second part of two part Julian date (TDB)

    Returns
    -------
    x : float or `np.ndarray`
        x coordinate of the CIP
    y : float or `np.ndarray`
        y coordinate of the CIP
    s : float or `np.ndarray`
        CIO locator, s
    """
    # classical NPB matrix, IAU 2006/2000A
    rpnb = erfa.pnm06a(jd1, jd2)
    # CIP X, Y coordinates from array
    x, y = erfa.bpn2xy(rpnb)
    # CIO locator, s
    s = erfa.s06(jd1, jd2, x, y)
    return x, y, s


def aticq(srepr, astrom):
    """
    A slightly modified version of the ERFA function ``eraAticq``.

    ``eraAticq`` performs the transformations between two coordinate systems,
    with the details of the transformation being encoded into the ``astrom`` array.

    There are two issues with the version of aticq in ERFA. Both are associated
    with the handling of light deflection.

    The companion function ``eraAtciqz`` is meant to be its inverse. However, this
    is not true for directions close to the Solar centre, since the light deflection
    calculations are numerically unstable and therefore not reversible.

    This version sidesteps that problem by artificially reducing the light deflection
    for directions which are within 90 arcseconds of the Sun's position. This is the
    same approach used by the ERFA functions above, except that they use a threshold of
    9 arcseconds.

    In addition, ERFA's aticq assumes a distant source, so there is no difference between
    the object-Sun vector and the observer-Sun vector. This can lead to errors of up to a
    few arcseconds in the worst case (e.g a Venus transit).

    Parameters
    ----------
    srepr : `~astropy.coordinates.SphericalRepresentation`
        Astrometric GCRS or CIRS position of object from observer
    astrom : eraASTROM array
        ERFA astrometry context, as produced by, e.g. ``eraApci13`` or ``eraApcs13``

    Returns
    -------
    rc : float or `~numpy.ndarray`
        Right Ascension in radians
    dc : float or `~numpy.ndarray`
        Declination in radians
    """
    # ignore parallax effects if no distance, or far away
    srepr_distance = srepr.distance
    ignore_distance = srepr_distance.unit == u.one

    # RA, Dec to cartesian unit vectors
    pos = erfa.s2c(srepr.lon.radian, srepr.lat.radian)

    # Bias-precession-nutation, giving GCRS proper direction.
    ppr = erfa.trxp(astrom["bpn"], pos)

    # Aberration, giving GCRS natural direction
    d = np.zeros_like(ppr)
    for _ in range(2):
        before = norm(ppr - d)
        after = erfa.ab(before, astrom["v"], astrom["em"], astrom["bm1"])
        d = after - before
    pnat = norm(ppr - d)

    # Light deflection by the Sun, giving BCRS coordinate direction
    d = np.zeros_like(pnat)
    for _ in range(5):
        before = norm(pnat - d)
        if ignore_distance:
            # No distance to object, assume a long way away
            q = before
        else:
            # Find BCRS direction of Sun to object.
            # astrom['eh'] and astrom['em'] contain Sun to observer unit vector,
            # and distance, respectively.
            eh = astrom["em"][..., np.newaxis] * astrom["eh"]
            # unit vector from Sun to object
            q = eh + srepr_distance[..., np.newaxis].to_value(u.au) * before
            sundist, q = erfa.pn(q)
            sundist = sundist[..., np.newaxis]
            # calculation above is extremely unstable very close to the sun
            # in these situations, default back to ldsun-style behaviour,
            # since this is reversible and drops to zero within stellar limb
            q = np.where(sundist > 1.0e-10, q, before)

        after = erfa.ld(1.0, before, q, astrom["eh"], astrom["em"], 1e-6)
        d = after - before
    pco = norm(pnat - d)

    # ICRS astrometric RA, Dec
    rc, dc = erfa.c2s(pco)
    return erfa.anp(rc), dc


def atciqz(srepr, astrom):
    """
    A slightly modified version of the ERFA function ``eraAtciqz``.

    ``eraAtciqz`` performs the transformations between two coordinate systems,
    with the details of the transformation being encoded into the ``astrom`` array.

    There are two issues with the version of atciqz in ERFA. Both are associated
    with the handling of light deflection.

    The companion function ``eraAticq`` is meant to be its inverse. However, this
    is not true for directions close to the Solar centre, since the light deflection
    calculations are numerically unstable and therefore not reversible.

    This version sidesteps that problem by artificially reducing the light deflection
    for directions which are within 90 arcseconds of the Sun's position. This is the
    same approach used by the ERFA functions above, except that they use a threshold of
    9 arcseconds.

    In addition, ERFA's atciqz assumes a distant source, so there is no difference between
    the object-Sun vector and the observer-Sun vector. This can lead to errors of up to a
    few arcseconds in the worst case (e.g a Venus transit).

    Parameters
    ----------
    srepr : `~astropy.coordinates.SphericalRepresentation`
        Astrometric ICRS position of object from observer
    astrom : eraASTROM array
        ERFA astrometry context, as produced by, e.g. ``eraApci13`` or ``eraApcs13``

    Returns
    -------
    ri : float or `~numpy.ndarray`
        Right Ascension in radians
    di : float or `~numpy.ndarray`
        Declination in radians
    """
    # ignore parallax effects if no distance, or far away
    srepr_distance = srepr.distance
    ignore_distance = srepr_distance.unit == u.one

    # BCRS coordinate direction (unit vector).
    pco = erfa.s2c(srepr.lon.radian, srepr.lat.radian)

    # Find BCRS direction of Sun to object
    if ignore_distance:
        # No distance to object, assume a long way away
        q = pco
    else:
        # Find BCRS direction of Sun to object.
        # astrom['eh'] and astrom['em'] contain Sun to observer unit vector,
        # and distance, respectively.
        eh = astrom["em"][..., np.newaxis] * astrom["eh"]
        # unit vector from Sun to object
        q = eh + srepr_distance[..., np.newaxis].to_value(u.au) * pco
        sundist, q = erfa.pn(q)
        sundist = sundist[..., np.newaxis]
        # calculation above is extremely unstable very close to the sun
        # in these situations, default back to ldsun-style behaviour,
        # since this is reversible and drops to zero within stellar limb
        q = np.where(sundist > 1.0e-10, q, pco)

    # Light deflection by the Sun, giving BCRS natural direction.
    pnat = erfa.ld(1.0, pco, q, astrom["eh"], astrom["em"], 1e-6)

    # Aberration, giving GCRS proper direction.
    ppr = erfa.ab(pnat, astrom["v"], astrom["em"], astrom["bm1"])

    # Bias-precession-nutation, giving CIRS proper direction.
    # Has no effect if matrix is identity matrix, in which case gives GCRS ppr.
    pi = erfa.rxp(astrom["bpn"], ppr)

    # CIRS (GCRS) RA, Dec
    ri, di = erfa.c2s(pi)
    return erfa.anp(ri), di


def prepare_earth_position_vel(time):
    """
    Get barycentric position and velocity, and heliocentric position of Earth.

    Parameters
    ----------
    time : `~astropy.time.Time`
        time at which to calculate position and velocity of Earth

    Returns
    -------
    earth_pv : `np.ndarray`
        Barycentric position and velocity of Earth, in au and au/day
    earth_helio : `np.ndarray`
        Heliocentric position of Earth in au
    """
    # this goes here to avoid circular import errors
    from astropy.coordinates.solar_system import (
        get_body_barycentric,
        get_body_barycentric_posvel,
        solar_system_ephemeris,
    )

    # get barycentric position and velocity of earth

    ephemeris = solar_system_ephemeris.get()

    # if we are using the builtin erfa based ephemeris,
    # we can use the fact that epv00 already provides all we need.
    # This avoids calling epv00 twice, once
    # in get_body_barycentric_posvel('earth') and once in
    # get_body_barycentric('sun')
    if ephemeris == "builtin":
        jd1, jd2 = get_jd12(time, "tdb")
        earth_pv_heliocentric, earth_pv = erfa.epv00(jd1, jd2)
        earth_heliocentric = earth_pv_heliocentric["p"]

    # all other ephemeris providers probably don't have a shortcut like this
    else:
        earth_p, earth_v = get_body_barycentric_posvel("earth", time)

        # get heliocentric position of earth, preparing it for passing to erfa.
        sun = get_body_barycentric("sun", time)
        earth_heliocentric = (earth_p - sun).get_xyz(xyz_axis=-1).to_value(u.au)

        # Also prepare earth_pv for passing to erfa, which wants it as
        # a structured dtype.
        earth_pv = pav2pv(
            earth_p.get_xyz(xyz_axis=-1).to_value(u.au),
            earth_v.get_xyz(xyz_axis=-1).to_value(u.au / u.d),
        )

    return earth_pv, earth_heliocentric


def get_offset_sun_from_barycenter(time, include_velocity=False, reverse=False):
    """
    Returns the offset of the Sun center from the solar-system barycenter (SSB).

    Parameters
    ----------
    time : `~astropy.time.Time`
        Time at which to calculate the offset
    include_velocity : `bool`
        If ``True``, attach the velocity as a differential.  Defaults to ``False``.
    reverse : `bool`
        If ``True``, return the offset of the barycenter from the Sun.  Defaults to ``False``.

    Returns
    -------
    `~astropy.coordinates.CartesianRepresentation`
        The offset
    """
    if include_velocity:
        # Import here to avoid a circular import
        from astropy.coordinates.solar_system import get_body_barycentric_posvel

        offset_pos, offset_vel = get_body_barycentric_posvel("sun", time)
        if reverse:
            offset_pos, offset_vel = -offset_pos, -offset_vel
        offset_vel = offset_vel.represent_as(CartesianDifferential)
        offset_pos = offset_pos.with_differentials(offset_vel)

    else:
        # Import here to avoid a circular import
        from astropy.coordinates.solar_system import get_body_barycentric

        offset_pos = get_body_barycentric("sun", time)
        if reverse:
            offset_pos = -offset_pos

    return offset_pos



# Licensed under a 3-clause BSD style license - see LICENSE.rst


# Note: This file includes code derived from pywcsgrid2
#
# This file contains Matplotlib transformation objects (e.g. from pixel to world
# coordinates, but also world-to-world).

import abc

import numpy as np
from matplotlib.path import Path
from matplotlib.transforms import Transform

from astropy import units as u
from astropy.coordinates import (
    BaseCoordinateFrame,
    SkyCoord,
    UnitSphericalRepresentation,
    frame_transform_graph,
)

__all__ = [
    "CoordinateTransform",
    "CurvedTransform",
    "Pixel2WorldTransform",
    "World2PixelTransform",
]


class CurvedTransform(Transform, metaclass=abc.ABCMeta):
    """
    Abstract base class for non-affine curved transforms.
    """

    input_dims = 2
    output_dims = 2
    is_separable = False

    def transform_path(self, path):
        """
        Transform a Matplotlib Path.

        Parameters
        ----------
        path : :class:`~matplotlib.path.Path`
            The path to transform

        Returns
        -------
        path : :class:`~matplotlib.path.Path`
            The resulting path
        """
        return Path(self.transform(path.vertices), path.codes)

    transform_path_non_affine = transform_path

    def transform(self, input):
        raise NotImplementedError("")

    def inverted(self):
        raise NotImplementedError("")


class CoordinateTransform(CurvedTransform):
    has_inverse = True

    def __init__(self, input_system, input_units, output_system, output_units):
        super().__init__()
        self._input_system_name = input_system
        self._output_system_name = output_system
        self._input_units = [u.Unit(unit) for unit in input_units]
        self._output_units = [u.Unit(unit) for unit in output_units]
        self.same_units = input_units == output_units

        if isinstance(self._input_system_name, str):
            frame_cls = frame_transform_graph.lookup_name(self._input_system_name)
            if frame_cls is None:
                raise ValueError(f"Frame {self._input_system_name} not found")
            else:
                self.input_system = frame_cls()
        elif isinstance(self._input_system_name, BaseCoordinateFrame):
            self.input_system = self._input_system_name
        else:
            raise TypeError(
                "input_system should be a WCS instance, string, or a coordinate frame"
                " instance"
            )

        if isinstance(self._output_system_name, str):
            frame_cls = frame_transform_graph.lookup_name(self._output_system_name)
            if frame_cls is None:
                raise ValueError(f"Frame {self._output_system_name} not found")
            else:
                self.output_system = frame_cls()

        elif isinstance(self._output_system_name, BaseCoordinateFrame):
            self.output_system = self._output_system_name
        else:
            raise TypeError(
                "output_system should be a WCS instance, string, or a coordinate frame"
                " instance"
            )

        if self.output_system == self.input_system:
            self.same_frames = True
        else:
            self.same_frames = False

    @property
    def same_frames(self):
        return self._same_frames

    @same_frames.setter
    def same_frames(self, same_frames):
        self._same_frames = same_frames

    def transform(self, input_coords):
        """
        Transform one set of coordinates to another.
        """
        if self.same_frames and self.same_units:
            return input_coords

        x_in = input_coords[:, 0] * u.Unit(self._input_units[0])
        y_in = input_coords[:, 1] * u.Unit(self._input_units[1])

        if self.same_frames:
            lon = x_in
            lat = y_in
        else:
            c_in = SkyCoord(
                UnitSphericalRepresentation(x_in, y_in), frame=self.input_system
            )

            # We often need to transform arrays that contain NaN values, and filtering
            # out the NaN values would have a performance hit, so instead we just pass
            # on all values and just ignore Numpy warnings
            with np.errstate(all="ignore"):
                c_out = c_in.transform_to(self.output_system)

            lon = c_out.spherical.lon
            lat = c_out.spherical.lat

        lon = lon.to_value(self._output_units[0])
        lat = lat.to_value(self._output_units[1])

        return np.concatenate((lon[:, np.newaxis], lat[:, np.newaxis]), axis=1)

    transform_non_affine = transform

    def inverted(self):
        """
        Return the inverse of the transform.
        """
        return CoordinateTransform(
            self._output_system_name,
            self._output_units,
            self._input_system_name,
            self._input_units,
        )


class World2PixelTransform(CurvedTransform, metaclass=abc.ABCMeta):
    """
    Base transformation from world to pixel coordinates.
    """

    has_inverse = True
    frame_in = None

    @property
    @abc.abstractmethod
    def input_dims(self):
        """
        The number of input world dimensions.
        """

    @abc.abstractmethod
    def transform(self, world):
        """
        Transform world to pixel coordinates. You should pass in a NxM array
        where N is the number of points to transform, and M is the number of
        dimensions. This then returns the (x, y) pixel coordinates
        as a Nx2 array.
        """

    @abc.abstractmethod
    def inverted(self):
        """
        Return the inverse of the transform.
        """


class Pixel2WorldTransform(CurvedTransform, metaclass=abc.ABCMeta):
    """
    Base transformation from pixel to world coordinates.
    """

    has_inverse = True
    frame_out = None

    @property
    @abc.abstractmethod
    def output_dims(self):
        """
        The number of output world dimensions.
        """

    @abc.abstractmethod
    def transform(self, pixel):
        """
        Transform pixel to world coordinates. You should pass in a Nx2 array
        of (x, y) pixel coordinates to transform to world coordinates. This
        will then return an NxM array where M is the number of dimensions.
        """

    @abc.abstractmethod
    def inverted(self):
        """
        Return the inverse of the transform.
        """



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

# NOTE: The configuration for the package, including the name, version, and
# other information are set in the pyproject.toml file.

import sys
from pathlib import Path

from setuptools import setup
from setuptools.command.editable_wheel import editable_wheel
from setuptools.command.install import install

from extension_helpers import get_extensions

ext_modules = get_extensions()


def install_stubs(build_lib, output_dir):
    sys.path.insert(0, build_lib)
    try:
        from astropy.units.typing_utils import create_stubs  # noqa: PLC0415

        create_stubs(Path(output_dir))
    finally:
        # Undo the path modification.
        sys.path.pop(0)


class InstallWithStubs(install):
    """Post-installation command for installation mode."""

    def run(self):
        super().run()
        install_stubs(self.build_lib, self.root)


class EditableInstallWithStubs(editable_wheel):
    """Post-installation command for editable_wheel mode."""

    def run(self):
        super().run()
        install_stubs(self.project_dir, self.project_dir)


# Specify the minimum version for the Numpy C-API
for ext in ext_modules:
    if ext.include_dirs and "numpy" in ext.include_dirs[0]:
        ext.define_macros.append(("NPY_TARGET_VERSION", "NPY_1_24_API_VERSION"))
        ext.define_macros.append(("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"))

setup(
    ext_modules=ext_modules,
    cmdclass={
        "install": InstallWithStubs,
        "editable_wheel": EditableInstallWithStubs,
    },
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# STDLIB
import typing

# THIRD PARTY
import numpy as np
import pytest

# LOCAL
from astropy import units as u

# list of pairs (target unit/physical type, input unit)
x_inputs = [
    (u.arcsec, u.deg),
    ("angle", u.deg),
    (u.kpc / u.Myr, u.km / u.s),
    ("speed", u.km / u.s),
    ([u.arcsec, u.km], u.deg),
    ([u.arcsec, u.km], u.km),  # multiple allowed
    (["angle", "length"], u.deg),
    (["angle", "length"], u.km),
]

y_inputs = [
    (u.m, u.km),
    (u.km, u.m),
    (u.arcsec, u.deg),
    ("angle", u.deg),
    (u.kpc / u.Myr, u.km / u.s),
    ("speed", u.km / u.s),
]


@pytest.fixture(scope="module", params=list(range(len(x_inputs))))
def x_input(request):
    return x_inputs[request.param]


@pytest.fixture(scope="module", params=list(range(len(y_inputs))))
def y_input(request):
    return y_inputs[request.param]


# ---- Tests that use the fixtures defined above ----


def test_args(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, y):
        return x, y

    x, y = myfunc_args(1 * x_unit, 1 * y_unit)

    assert isinstance(x, u.Quantity)
    assert isinstance(y, u.Quantity)

    assert x.unit == x_unit
    assert y.unit == y_unit


def test_args_nonquantity(x_input):
    x_target, x_unit = x_input

    @u.quantity_input(x=x_target)
    def myfunc_args(x, y):
        return x, y

    x, y = myfunc_args(1 * x_unit, 100)

    assert isinstance(x, u.Quantity)
    assert isinstance(y, int)

    assert x.unit == x_unit


def test_wrong_unit(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, y):
        return x, y

    with pytest.raises(
        u.UnitsError,
        match=(
            "Argument 'y' to function 'myfunc_args' must be in units "
            f"convertible to '{str(y_target)}'."
        ),
    ):
        x, y = myfunc_args(1 * x_unit, 100 * u.Joule)  # has to be an unspecified unit


def test_wrong_unit_annotated(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input
    def myfunc_args(x: x_target, y: y_target):
        return x, y

    with pytest.raises(u.UnitsError, match="Argument 'y' to function 'myfunc_args'"):
        x, y = myfunc_args(1 * x_unit, 100 * u.Joule)  # has to be an unspecified unit


def test_not_quantity(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, y):
        return x, y

    with pytest.raises(
        TypeError,
        match=(
            "Argument 'y' to function 'myfunc_args' has no 'unit' attribute. "
            "You should pass in an astropy Quantity instead."
        ),
    ):
        x, y = myfunc_args(1 * x_unit, 100)


def test_not_quantity_annotated(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input
    def myfunc_args(x: x_target, y: y_target):
        return x, y

    with pytest.raises(
        TypeError,
        match=(
            "Argument 'y' to function 'myfunc_args' has no 'unit' attribute. "
            "You should pass in an astropy Quantity instead."
        ),
    ):
        x, y = myfunc_args(1 * x_unit, 100)


def test_kwargs(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, my_arg, y=1 * y_unit):
        return x, my_arg, y

    x, my_arg, y = myfunc_args(1 * x_unit, 100, y=100 * y_unit)

    assert isinstance(x, u.Quantity)
    assert isinstance(my_arg, int)
    assert isinstance(y, u.Quantity)

    assert y.unit == y_unit


def test_unused_kwargs(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, my_arg1, y=y_unit, my_arg2=1000):
        return x, my_arg1, y, my_arg2

    x, my_arg1, y, my_arg2 = myfunc_args(1 * x_unit, 100, y=100 * y_unit, my_arg2=10)

    assert isinstance(x, u.Quantity)
    assert isinstance(my_arg1, int)
    assert isinstance(y, u.Quantity)
    assert isinstance(my_arg2, int)

    assert y.unit == y_unit
    assert my_arg2 == 10


def test_kwarg_wrong_unit(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, y=10 * y_unit):
        return x, y

    with pytest.raises(
        u.UnitsError,
        match=(
            "Argument 'y' to function 'myfunc_args' must be in units "
            f"convertible to '{str(y_target)}'."
        ),
    ):
        x, y = myfunc_args(1 * x_unit, y=100 * u.Joule)


def test_kwarg_not_quantity(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, y=10 * y_unit):
        return x, y

    with pytest.raises(
        TypeError,
        match=(
            "Argument 'y' to function 'myfunc_args' has no 'unit' attribute. "
            "You should pass in an astropy Quantity instead."
        ),
    ):
        x, y = myfunc_args(1 * x_unit, y=100)


def test_kwarg_default(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, y=10 * y_unit):
        return x, y

    x, y = myfunc_args(1 * x_unit)

    assert isinstance(x, u.Quantity)
    assert isinstance(y, u.Quantity)

    assert x.unit == x_unit
    assert y.unit == y_unit


def test_kwargs_input(x_input, y_input):
    x_target, x_unit = x_input
    y_target, y_unit = y_input

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x=1 * x_unit, y=1 * y_unit):
        return x, y

    kwargs = {"x": 10 * x_unit, "y": 10 * y_unit}
    x, y = myfunc_args(**kwargs)

    assert isinstance(x, u.Quantity)
    assert isinstance(y, u.Quantity)

    assert x.unit == x_unit
    assert y.unit == y_unit


def test_kwargs_extra(x_input):
    x_target, x_unit = x_input

    @u.quantity_input(x=x_target)
    def myfunc_args(x, **kwargs):
        return x

    x = myfunc_args(1 * x_unit)

    assert isinstance(x, u.Quantity)

    assert x.unit == x_unit


# ---- Tests that don't used the fixtures ----


@pytest.mark.parametrize("x_unit,y_unit", [(u.arcsec, u.eV), ("angle", "energy")])
def test_arg_equivalencies(x_unit, y_unit):
    @u.quantity_input(x=x_unit, y=y_unit, equivalencies=u.mass_energy())
    def myfunc_args(x, y):
        return x, y + (10 * u.J)  # Add an energy to check equiv is working

    x, y = myfunc_args(1 * u.arcsec, 100 * u.gram)

    assert isinstance(x, u.Quantity)
    assert isinstance(y, u.Quantity)

    assert x.unit == u.arcsec
    assert y.unit == u.gram


@pytest.mark.parametrize("x_unit,energy_unit", [(u.arcsec, u.eV), ("angle", "energy")])
def test_kwarg_equivalencies(x_unit, energy_unit):
    @u.quantity_input(x=x_unit, energy=energy_unit, equivalencies=u.mass_energy())
    def myfunc_args(x, energy=10 * u.eV):
        return x, energy + (10 * u.J)  # Add an energy to check equiv is working

    x, energy = myfunc_args(1 * u.arcsec, 100 * u.gram)

    assert isinstance(x, u.Quantity)
    assert isinstance(energy, u.Quantity)

    assert x.unit == u.arcsec
    assert energy.unit == u.gram


def test_no_equivalent():
    class test_unit:
        pass

    class test_quantity:
        unit = test_unit()

    @u.quantity_input(x=u.arcsec)
    def myfunc_args(x):
        return x

    with pytest.raises(
        TypeError,
        match=(
            "Argument 'x' to function 'myfunc_args' has a 'unit' attribute without an"
            " 'is_equivalent' method. You should pass in an astropy Quantity instead."
        ),
    ):
        x, y = myfunc_args(test_quantity())


def test_kwarg_invalid_physical_type():
    @u.quantity_input(x="angle", y="africanswallow")
    def myfunc_args(x, y=10 * u.deg):
        return x, y

    with pytest.raises(
        ValueError, match="Invalid unit or physical type 'africanswallow'."
    ):
        x, y = myfunc_args(1 * u.arcsec, y=100 * u.deg)


def test_default_value_check():
    x_target = u.deg
    x_unit = u.arcsec

    with pytest.raises(TypeError):

        @u.quantity_input(x=x_target)
        def myfunc_args(x=1.0):
            return x

        x = myfunc_args()

    x = myfunc_args(1 * x_unit)
    assert isinstance(x, u.Quantity)
    assert x.unit == x_unit


def test_str_unit_typo():
    @u.quantity_input
    def myfunc_args(x: "kilograam"):  # noqa: F821
        return x

    with pytest.raises(ValueError):
        result = myfunc_args(u.kg)


class TestTypeAnnotations:
    @pytest.mark.parametrize(
        "annot",
        [u.m, u.Quantity[u.m], u.Quantity[u.m, "more"]],
    )  # Note: parametrization is done even if test class is skipped
    def test_single_annotation_unit(self, annot):
        """Try a variety of valid annotations."""

        @u.quantity_input
        def myfunc_args(x: annot, y: str):
            return x, y

        i_q, i_str = 2 * u.m, "cool string"
        o_q, o_str = myfunc_args(i_q, i_str)
        assert o_q == i_q
        assert o_str == i_str


def test_args_None():
    x_target = u.deg
    x_unit = u.arcsec
    y_target = u.km
    y_unit = u.kpc

    @u.quantity_input(x=[x_target, None], y=[None, y_target])
    def myfunc_args(x, y):
        return x, y

    x, y = myfunc_args(1 * x_unit, None)
    assert isinstance(x, u.Quantity)
    assert x.unit == x_unit
    assert y is None

    x, y = myfunc_args(None, 1 * y_unit)
    assert isinstance(y, u.Quantity)
    assert y.unit == y_unit
    assert x is None


def test_args_None_kwarg():
    x_target = u.deg
    x_unit = u.arcsec
    y_target = u.km

    @u.quantity_input(x=x_target, y=y_target)
    def myfunc_args(x, y=None):
        return x, y

    x, y = myfunc_args(1 * x_unit)
    assert isinstance(x, u.Quantity)
    assert x.unit == x_unit
    assert y is None

    x, y = myfunc_args(1 * x_unit, None)
    assert isinstance(x, u.Quantity)
    assert x.unit == x_unit
    assert y is None

    with pytest.raises(TypeError):
        x, y = myfunc_args(None, None)


@pytest.mark.parametrize("val", [1.0, 1, np.arange(10), np.arange(10.0)])
def test_allow_dimensionless_numeric(val):
    """
    When dimensionless_unscaled is an allowed unit, numbers and numeric numpy
    arrays are allowed through
    """

    @u.quantity_input(velocity=[u.km / u.s, u.dimensionless_unscaled])
    def myfunc(velocity):
        return velocity

    assert np.all(myfunc(val) == val)


@pytest.mark.parametrize("val", [1.0, 1, np.arange(10), np.arange(10.0)])
def test_allow_dimensionless_numeric_strict(val):
    """
    When dimensionless_unscaled is an allowed unit, but we are being strict,
    don't allow numbers and numeric numpy arrays through
    """

    @u.quantity_input(
        velocity=[u.km / u.s, u.dimensionless_unscaled], strict_dimensionless=True
    )
    def myfunc(velocity):
        return velocity

    with pytest.raises(TypeError):
        assert myfunc(val)


@pytest.mark.parametrize("val", [1 * u.deg, [1, 2, 3] * u.m])
def test_dimensionless_with_nondimensionless_input(val):
    """
    When dimensionless_unscaled is the only allowed unit, don't let input with
    non-dimensionless units through
    """

    @u.quantity_input(x=u.dimensionless_unscaled)
    def myfunc(x):
        return x

    with pytest.raises(u.UnitsError):
        myfunc(val)


def test_annotated_not_quantity():
    """Test when annotation looks like a Quantity[X], but isn't."""

    @u.quantity_input()
    def myfunc(x: typing.Annotated[object, u.m]):
        return x

    # nothing happens when wrong unit is passed
    assert myfunc(1) == 1
    assert myfunc(1 * u.m) == 1 * u.m
    assert myfunc(1 * u.s) == 1 * u.s


def test_annotated_not_unit():
    """Test when annotation looks like a Quantity[X], but the unit's wrong."""

    @u.quantity_input()
    def myfunc(x: typing.Annotated[u.Quantity, object()]):
        return x

    # nothing happens when wrong unit is passed
    assert myfunc(1) == 1
    assert myfunc(1 * u.m) == 1 * u.m
    assert myfunc(1 * u.s) == 1 * u.s



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Astropy cosmology core module.

.. deprecated:: 7.1

    This module is deprecated and will be removed in a future version. All the public
    classes and functions have been and will continue to be available in the
    :mod:`~astropy.cosmology` module.

"""

__all__ = ["Cosmology", "CosmologyError", "FlatCosmologyMixin"]  # noqa: F822


import sys
import warnings
from typing import Any


def __getattr__(name: str) -> Any:
    if name not in __all__ + ["_COSMOLOGY_CLASSES", "dataclass_decorator"]:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}.")

    from ._src import core

    obj = getattr(core, name)

    setattr(sys.modules[__name__], name, obj)

    warnings.warn(
        "The module `astropy.cosmology.core` is deprecated since v7.1 and will be "
        "removed in a future version. Import from `astropy.cosmology` instead.",
        category=DeprecationWarning,
    )

    return obj



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from math import ceil

import numpy as np

from astropy.utils.exceptions import AstropyUserWarning

__all__ = ["_validate_tile_shape"]


def _validate_tile_shape(*, tile_shape, compression_type, image_header):
    naxis = image_header["NAXIS"]

    if not tile_shape:
        tile_shape = []
    elif len(tile_shape) != naxis:
        warnings.warn(
            "Provided tile size not appropriate for the data.  "
            "Default tile size will be used.",
            AstropyUserWarning,
        )
        tile_shape = []
    else:
        tile_shape = list(tile_shape)

    # Set default tile dimensions for HCOMPRESS_1

    if compression_type == "HCOMPRESS_1":
        if image_header["NAXIS1"] < 4 or image_header["NAXIS2"] < 4:
            raise ValueError("Hcompress minimum image dimension is 4 pixels")
        elif tile_shape:
            if tile_shape[-1] < 4 or tile_shape[-2] < 4:
                # user specified tile size is too small
                raise ValueError("Hcompress minimum tile dimension is 4 pixels")
            major_dims = len([ts for ts in tile_shape if ts > 1])
            if major_dims > 2:
                raise ValueError(
                    "HCOMPRESS can only support 2-dimensional tile sizes."
                    "All but two of the tile_shape dimensions must be set "
                    "to 1."
                )

        if tile_shape and (tile_shape[-1] == 0 and tile_shape[-2] == 0):
            # compress the whole image as a single tile
            tile_shape[-1] = image_header["NAXIS1"]
            tile_shape[-2] = image_header["NAXIS2"]

            for i in range(2, naxis):
                # set all higher tile dimensions = 1
                tile_shape[i] = 1
        elif not tile_shape:
            # The Hcompress algorithm is inherently 2D in nature, so the
            # row by row tiling that is used for other compression
            # algorithms is not appropriate.  If the image has less than 30
            # rows, then the entire image will be compressed as a single
            # tile.  Otherwise the tiles will consist of 16 rows of the
            # image.  This keeps the tiles to a reasonable size, and it
            # also includes enough rows to allow good compression
            # efficiency.  It the last tile of the image happens to contain
            # less than 4 rows, then find another tile size with between 14
            # and 30 rows (preferably even), so that the last tile has at
            # least 4 rows.

            # 1st tile dimension is the row length of the image
            tile_shape = [image_header["NAXIS1"]]

            if image_header["NAXIS2"] <= 30:
                tile_shape.insert(0, image_header["NAXIS1"])
            else:
                # look for another good tile dimension
                naxis2 = image_header["NAXIS2"]
                for dim in [16, 24, 20, 30, 28, 26, 22, 18, 14]:
                    if naxis2 % dim == 0 or naxis2 % dim > 3:
                        tile_shape.insert(0, dim)
                        break
                else:
                    tile_shape.insert(0, 17)

            for i in range(2, naxis):
                # set all higher tile dimensions = 1
                tile_shape.insert(0, 1)

        # check if requested tile size causes the last tile to have
        # less than 4 pixels
        remain = image_header["NAXIS1"] % tile_shape[-1]  # 1st dimen
        if 0 < remain < 4:
            raise ValueError("Last tile along 1st dimension has less than 4 pixels")

        remain = image_header["NAXIS2"] % tile_shape[-2]  # 2nd dimen
        if 0 < remain < 4:
            raise ValueError("Last tile along 2nd dimension has less than 4 pixels")

    if len(tile_shape) == 0 and image_header["NAXIS"] > 0:
        tile_shape = [1] * (naxis - 1) + [image_header["NAXIS1"]]

    return tuple(tile_shape)


def _n_tiles(data_shape, tile_shape):
    return [ceil(d / t) for d, t in zip(data_shape, tile_shape)]


def _iter_array_tiles(
    data_shape, tile_shape, first_tile_index=None, last_tile_index=None
):
    ndim = len(tile_shape)

    n_tiles = _n_tiles(data_shape, tile_shape)

    if first_tile_index is None:
        first_tile_index = (0,) * ndim

    if last_tile_index is None:
        last_tile_index = tuple(n - 1 for n in n_tiles)

    tile_index = list(first_tile_index)

    while True:
        tile_slices = tuple(
            slice(
                (tile_index[idx] - first_tile_index[idx]) * tile_shape[idx],
                (tile_index[idx] - first_tile_index[idx] + 1) * tile_shape[idx],
            )
            for idx in range(ndim)
        )

        row_index = tile_index[0]
        for dim in range(1, ndim):
            row_index = tile_index[dim] + row_index * n_tiles[dim]

        yield row_index, tile_slices

        tile_index[-1] += 1

        for idx in range(ndim - 1, 0, -1):
            if tile_index[idx] > last_tile_index[idx]:
                tile_index[idx] = first_tile_index[idx]
                tile_index[idx - 1] += 1

        if tile_index[0] > last_tile_index[0]:
            break


def _tile_shape(header):
    return np.array(
        [header[f"ZTILE{idx}"] for idx in range(header["ZNAXIS"], 0, -1)], dtype=int
    )


def _data_shape(header):
    return np.array(
        [header[f"ZNAXIS{idx}"] for idx in range(header["ZNAXIS"], 0, -1)], dtype=int
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import sys
from collections import defaultdict
from os.path import join

from setuptools import Extension

from extension_helpers import pkg_config


def get_extensions(build_type="release"):
    XML_DIR = "astropy/utils/xml/src"

    cfg = defaultdict(list)
    cfg["sources"] = [join(XML_DIR, "iterparse.c")]

    if int(os.environ.get("ASTROPY_USE_SYSTEM_EXPAT", "0")) or int(
        os.environ.get("ASTROPY_USE_SYSTEM_ALL", "0")
    ):
        for k, v in pkg_config(["expat"], ["expat"]).items():
            cfg[k].extend(v)
    else:
        EXPAT_DIR = "cextern/expat/lib"
        cfg["sources"].extend(
            [
                join(EXPAT_DIR, fn)
                for fn in ["xmlparse.c", "xmlrole.c", "xmltok.c", "xmltok_impl.c"]
            ]
        )
        cfg["include_dirs"].extend([XML_DIR, EXPAT_DIR])
        if sys.platform.startswith("linux"):
            # This is to ensure we only export the Python entry point
            # symbols and the linker won't try to use the system expat in
            # place of ours.
            cfg["extra_link_args"].extend(
                [f"-Wl,--version-script={join(XML_DIR, 'iterparse.map')}"]
            )
        cfg["define_macros"].append(("HAVE_EXPAT_CONFIG_H", 1))
        if sys.byteorder == "big":
            cfg["define_macros"].append(("BYTEORDER", "4321"))
        else:
            cfg["define_macros"].append(("BYTEORDER", "1234"))
        if sys.platform != "win32":
            cfg["define_macros"].append(("HAVE_UNISTD_H", None))

    return [Extension("astropy.utils.xml._iterparser", **cfg)]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# ruff: noqa: PYI024

# namedtuple is needed for find_mod_objs so it can have a non-local module
from collections import namedtuple
from unittest import mock

import pytest
import yaml

from astropy.utils import introspection
from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.utils.introspection import (
    find_current_module,
    find_mod_objs,
    isinstancemethod,
    minversion,
    resolve_name,
)


def test_pkg_finder():
    """
    Tests that the `find_current_module` function works. Note that
    this also implicitly tests compat.misc._patched_getmodule
    """
    mod1 = "astropy.utils.introspection"
    mod2 = "astropy.utils.tests.test_introspection"
    mod3 = "astropy.utils.tests.test_introspection"
    assert find_current_module(0).__name__ == mod1
    assert find_current_module(1).__name__ == mod2
    assert find_current_module(0, True).__name__ == mod3


def test_find_current_mod():
    from sys import getrecursionlimit

    thismodnm = __name__

    assert find_current_module(0) is introspection
    assert find_current_module(1).__name__ == thismodnm
    assert find_current_module(getrecursionlimit() + 1) is None

    assert find_current_module(0, True).__name__ == thismodnm
    assert find_current_module(0, [introspection]).__name__ == thismodnm
    assert find_current_module(0, ["astropy.utils.introspection"]).__name__ == thismodnm

    with pytest.raises(ImportError):
        find_current_module(0, ["faddfdsasewrweriopunjlfiurrhujnkflgwhu"])


def test_find_mod_objs():
    deprecation_message = (
        "^The find_mod_objs function is deprecated and may be removed in "
        r"a future version\.$"
    )
    with pytest.warns(AstropyDeprecationWarning, match=deprecation_message):
        lnms, fqns, objs = find_mod_objs("astropy")

    # this import  is after the above call intentionally to make sure
    # find_mod_objs properly imports astropy on its own
    import astropy

    # just check for astropy.conf ... other things might be added, so we
    # shouldn't check that it's the only thing
    assert "conf" in lnms
    assert astropy.conf in objs

    with pytest.warns(AstropyDeprecationWarning, match=deprecation_message):
        lnms, fqns, objs = find_mod_objs(__name__, onlylocals=False)
    assert "namedtuple" in lnms
    assert "collections.namedtuple" in fqns
    assert namedtuple in objs

    with pytest.warns(AstropyDeprecationWarning, match=deprecation_message):
        lnms, fqns, objs = find_mod_objs(__name__, onlylocals=True)
    assert "namedtuple" not in lnms
    assert "collections.namedtuple" not in fqns
    assert namedtuple not in objs


def test_minversion():
    import numpy as np

    good_versions = ["1.16", "1.16.1", "1.16.0.dev", "1.16dev"]
    bad_versions = ["100000", "100000.2rc1"]
    for version in good_versions:
        assert minversion(np, version)
        assert minversion("numpy", version)
    for version in bad_versions:
        assert not minversion(np, version)
        assert not minversion("numpy", version)

    assert minversion(yaml, "3.1")
    assert minversion("yaml", "3.1")


def test_find_current_module_bundle():
    """
    Tests that the `find_current_module` function would work if used inside
    an application bundle. Since we can't test this directly, we test what
    would happen if inspect.getmodule returned `None`, which is what happens
    inside PyInstaller and py2app bundles.
    """
    with mock.patch("inspect.getmodule", return_value=None):
        mod1 = "astropy.utils.introspection"
        mod2 = "astropy.utils.tests.test_introspection"
        mod3 = "astropy.utils.tests.test_introspection"
        assert find_current_module(0).__name__ == mod1
        assert find_current_module(1).__name__ == mod2
        assert find_current_module(0, True).__name__ == mod3


def test_deprecated_isinstancemethod():
    class MetaClass(type):
        def a_classmethod(cls):
            pass

    class MyClass(metaclass=MetaClass):
        def an_instancemethod(self):
            pass

        @classmethod
        def another_classmethod(cls):
            pass

        @staticmethod
        def a_staticmethod():
            pass

    deprecation_message = (
        "^The isinstancemethod function is deprecated and may be removed in "
        r"a future version\.$"
    )
    with pytest.warns(AstropyDeprecationWarning, match=deprecation_message):
        assert isinstancemethod(MyClass, MyClass.a_classmethod) is False
    with pytest.warns(AstropyDeprecationWarning, match=deprecation_message):
        assert isinstancemethod(MyClass, MyClass.another_classmethod) is False
    with pytest.warns(AstropyDeprecationWarning, match=deprecation_message):
        assert isinstancemethod(MyClass, MyClass.a_staticmethod) is False
    with pytest.warns(AstropyDeprecationWarning, match=deprecation_message):
        assert isinstancemethod(MyClass, MyClass.an_instancemethod) is True


def test_resolve_name_deprecation():
    with pytest.warns(
        AstropyDeprecationWarning,
        match=(
            "^The resolve_name function is deprecated and may be removed in a future "
            r"version\.\n        Use importlib \(e\.g\. importlib\.import_module for "
            r"modules\) instead\.$"
        ),
    ):
        assert resolve_name("astropy.utils.introspection") is introspection



# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__: tuple[str, ...] = ()  # nothing is publicly scoped

import functools
import operator
from dataclasses import Field
from typing import TypeGuard

from .core import Parameter


def is_parameter_or_field(obj: object, /) -> TypeGuard[Parameter | Field[Parameter]]:
    """Return if object is a Parameter or dataclass field thereof."""
    return isinstance(obj, Parameter) or (
        isinstance(obj, Field) and isinstance(obj.default, Parameter)
    )


def all_parameters(obj: object, /) -> dict[str, Parameter]:
    """Get all `Parameter` fields of an object.

    Returns all fields of the object, including those not yet finalized in the class, if
    it's still under construction, e.g. in ``__init_subclass__``.
    """
    cls = obj if isinstance(obj, type) else obj.__class__
    all_cls_vars = functools.reduce(operator.__or__, map(vars, cls.mro()[::-1]))

    return {
        k: (v if isinstance(v, Parameter) else v.default)
        for k, v in all_cls_vars.items()
        if is_parameter_or_field(v)
    }



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# pylint: disable=invalid-name
import re
from inspect import Parameter

import numpy as np
import pytest

from astropy.modeling.utils import (
    _SpecialOperatorsDict,
    _validate_domain_window,
    get_inputs_and_params,
    poly_map_domain,
    quantity_asanyarray,
)
from astropy.units import Quantity


def test_poly_map_domain():
    oldx = np.array([1, 2, 3, 4])

    # test shift/scale
    assert (poly_map_domain(oldx, (-4, 4), (-3, 3)) == [0.75, 1.5, 2.25, 3]).all()

    # errors
    MESSAGE = r'Expected "domain" and "window" to be a tuple of size 2'
    with pytest.raises(ValueError, match=MESSAGE):
        poly_map_domain(oldx, (-4,), (-3, 3))
    with pytest.raises(ValueError, match=MESSAGE):
        poly_map_domain(oldx, (-4, 4, -4), (-3, 3))
    with pytest.raises(ValueError, match=MESSAGE):
        poly_map_domain(oldx, (-4, 4), (-3,))
    with pytest.raises(ValueError, match=MESSAGE):
        poly_map_domain(oldx, (-4, 4), (-3, 3, -3))


def test__validate_domain_window():
    # Test if None
    assert _validate_domain_window(None) is None

    # Test normal
    assert _validate_domain_window((-2, 2)) == (-2, 2)
    assert _validate_domain_window([-2, 2]) == (-2, 2)
    assert _validate_domain_window(np.array([-2, 2])) == (-2, 2)

    # Test error
    MESSAGE = r"domain and window should be tuples of size 2"
    with pytest.raises(ValueError, match=MESSAGE):
        _validate_domain_window((-2, 2, -2))
    with pytest.raises(ValueError, match=MESSAGE):
        _validate_domain_window((-2,))
    with pytest.raises(ValueError, match=MESSAGE):
        _validate_domain_window([-2])
    with pytest.raises(ValueError, match=MESSAGE):
        _validate_domain_window(np.array([-2]))
    with pytest.raises(ValueError, match=MESSAGE):
        _validate_domain_window(-2)


def test_get_inputs_and_params():
    # test normal
    def func1(input0, input1, param0=5, param1=7):
        pass

    inputs, params = get_inputs_and_params(func1)
    for index, _input in enumerate(inputs):
        assert isinstance(_input, Parameter)
        assert _input.name == f"input{index}"
        assert _input.kind == _input.POSITIONAL_OR_KEYWORD
        assert _input.default == Parameter.empty
    default = [5, 7]
    for index, param in enumerate(params):
        assert isinstance(param, Parameter)
        assert param.name == f"param{index}"
        assert param.kind == param.POSITIONAL_OR_KEYWORD
        assert param.default == default[index]

    # Error
    MESSAGE = re.escape("Signature must not have *args or **kwargs")

    def func2(input0, input1, *args, param0=5, param1=7):
        pass

    def func3(input0, input1, param0=5, param1=7, **kwargs):
        pass

    with pytest.raises(ValueError, match=MESSAGE):
        get_inputs_and_params(func2)
    with pytest.raises(ValueError, match=MESSAGE):
        get_inputs_and_params(func3)


class Test_SpecialOperatorsDict:
    def setup_method(self):
        self.key = "test"
        self.val = "value"

    def test__set_value(self):
        special_operators = _SpecialOperatorsDict()
        assert self.key not in special_operators

        special_operators._set_value(self.key, self.val)
        assert self.key in special_operators
        assert special_operators[self.key] == self.val

        with pytest.raises(ValueError, match='Special operator "test" already exists'):
            special_operators._set_value(self.key, self.val)

    def test___setitem__(self):
        special_operators = _SpecialOperatorsDict()
        assert self.key not in special_operators

        with pytest.deprecated_call():
            special_operators[self.key] = self.val
        assert self.key in special_operators
        assert special_operators[self.key] == self.val

    def test__SpecialOperatorsDict__get_unique_id(self):
        special_operators = _SpecialOperatorsDict()
        assert special_operators._unique_id == 0

        assert special_operators._get_unique_id() == 1
        assert special_operators._unique_id == 1

        assert special_operators._get_unique_id() == 2
        assert special_operators._unique_id == 2

        assert special_operators._get_unique_id() == 3
        assert special_operators._unique_id == 3

    def test__SpecialOperatorsDict_add(self):
        special_operators = _SpecialOperatorsDict()

        operator_name = "test"
        operator = "operator"

        key0 = special_operators.add(operator_name, operator)
        assert key0 == (operator_name, special_operators._unique_id)
        assert key0 in special_operators
        assert special_operators[key0] == operator

        key1 = special_operators.add(operator_name, operator)
        assert key1 == (operator_name, special_operators._unique_id)
        assert key1 in special_operators
        assert special_operators[key1] == operator

        assert key0 != key1


def test_quantity_asanyarray():
    array_of_quantities = [Quantity(1), Quantity(2), Quantity(3)]
    quantity_array = quantity_asanyarray(array_of_quantities)
    assert isinstance(quantity_array, Quantity)
    assert np.issubdtype(quantity_array.dtype, np.inexact)

    array_of_integers = [1, 2, 3]
    np_array = quantity_asanyarray(array_of_integers)
    assert isinstance(np_array, np.ndarray)
    assert np.issubdtype(np_array.dtype, np.integer)

    np_array = quantity_asanyarray(array_of_integers, dtype=np.inexact)
    assert np.issubdtype(np_array.dtype, np.inexact)
